import numpy as np
import pandas as pd
import pytest
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import make_pipeline

from skorecard import Skorecard
from skorecard.bucketers import DecisionTreeBucketer, OrdinalCategoricalBucketer
from skorecard.rescale import ScoreCardPoints, calibrate_to_master_scale


@pytest.fixture()
def probas():
    """Proba array fixture."""
    return np.array([0.1 * i for i in range(1, 10)])


def test_master_scale_calibration(probas):
    """Test that the master scale calibrations follows the expected logic."""
    scores = calibrate_to_master_scale(probas, pdo=20, ref_odds=1, ref_score=100)
    # add odds 1:1, the score is 100. This are probas of 50%

    # When proba is 0.2, the odds are 4:1. This means that the score should have increased by twice the pdo
    assert scores[1] == 140

    # When proba is 0.8, the odds are 1:4. This means that the score should have decreased by twice the pdo
    assert scores[7] == 60

    scores = calibrate_to_master_scale(probas, pdo=30, ref_odds=1, ref_score=100)
    # add odds 1:1, the score is 100. This are probas of 50%

    # When proba is 0.2, the odds are 4:1. This means that the score should have increased by twice the pdo
    assert scores[1] == 160

    # When proba is 0.8, the odds are 1:4. This means that the score should have decreased by twice the pdo
    assert scores[7] == 40


def test_scorecard_rescaling(df):
    """Test scorecard rescaling."""
    features = ["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "MARRIAGE"]

    X = df[features]
    y = df["default"]

    scorecard_model = Skorecard(
        bucketing=make_pipeline(
            DecisionTreeBucketer(variables=features[:2]),
            OrdinalCategoricalBucketer(variables=features[2:]),
        )
    )
    with pytest.raises(NotFittedError):
        ScoreCardPoints(skorecard_model=scorecard_model, pdo=25, ref_score=400, ref_odds=20)

    scorecard_model.fit(X, y)

    scorecard = ScoreCardPoints(
        skorecard_model=scorecard_model, pdo=25, ref_score=400, ref_odds=20
    ).get_scorecard_points()

    # Contains all the features and the intercept
    assert len(scorecard.loc[:, "feature"].unique().tolist()) == len(features) + 1

    # The intercept must have 0 points
    assert scorecard.loc[scorecard["feature"] == "Intercept", "Points"].values[0] == 0

    # Make sure transform works
    sc = ScoreCardPoints(skorecard_model=scorecard_model, pdo=25, ref_score=400, ref_odds=20)
    assert isinstance(sc.transform(X), pd.DataFrame)



import numpy as np
import pandas as pd
from sklearn.metrics import make_scorer


def woe_1d(X, y, epsilon=0.00001):
    """Compute the weight of evidence on a 1-dimensional array.

    Args:
        X (np.array): bucketed dataframe
        y (np.array): target
        epsilon (float): Amount to be added to relative counts in order to avoid division by zero in the WOE
            calculation.

    Returns: (dictionary)
        - bins: indices of unique values of X
        - woe_values: calculated weight of evidence for every unique bin
        - counts_0: count of entries per bin where y==0
        - counts_1: count of entries per bin where y==1
    """
    # Make sure y has the right number of rows.
    if y.shape[0] != X.shape[0]:
        raise ValueError(f"y has {y.shape[0]}, but expected {X.shape[0]}")

    # Make sure y is a pd.Series so we can reset its index.
    if not isinstance(y, pd.Series):
        y = pd.Series(y)

    X = X.reset_index(drop=True)
    y = y.reset_index(drop=True)

    # Ensure classes in y start at zero.
    y = y - min(y)

    df = pd.concat([X, y], axis=1, ignore_index=True)
    df.columns = ["feat", "target"]

    total_pos = df["target"].sum()
    total_neg = df.shape[0] - total_pos
    df["non_target"] = np.where(df["target"] == 1, 0, 1)

    pos = (df.groupby(["feat"])["target"].sum() / total_pos) + epsilon
    neg = (df.groupby(["feat"])["non_target"].sum() / total_neg) + epsilon

    # Make sure to give informative error when dividing by zero error occurs
    msg = """
    One of the unique values in X has no occurrences of the %s class.
    Set epsilon to a very small value, or use a more coarse binning.
    """
    if any(neg == 0):
        raise ZeroDivisionError(msg % "negative")
    if any(pos == 0):
        raise ZeroDivisionError(msg % "positive")

    t = pd.concat([pos, neg], axis=1)
    t["woe"] = np.log(t["non_target"] / t["target"])

    if t["woe"].isnull().any():
        msg = "Woe Calculation produced NaNs! "
        msg += "Perhaps check your target distribution contains more than 1 class."
        raise ValueError(msg)

    return t


def _IV_score(y_test, y_pred, epsilon=0.0001, digits=None):
    """Using the unique values in y_pred, calculates the information value for the specific np.array.

    Args:
        y_test: (np.array), binary features, target
        y_pred: (np.array), predictions, indices of the buckets where the IV should be computed
        epsilon (float): Amount to be added to relative counts in order to avoid division by zero in the WOE
            calculation.
        digits: (int): number of significant decimal digits in the IV calculation

    Returns:
        iv (float): information value

    """
    df = woe_1d(y_pred, y_test, epsilon=epsilon)

    iv = ((df["non_target"] - df["target"]) * df["woe"]).sum()

    if digits:
        assert isinstance(digits, int), f"Digits must be an integer! Passed a {type(int)} variable"

        iv = np.round(iv, digits)
    return iv


@make_scorer
def IV_scorer(y_test, y_pred):
    """Decorated version, makes the IV score usable for sklearn grid search pipelines.

    Using the unique values in y_pred, calculates the information value for the specific np.array.

    Args:
        y_test: (np.array), binary features, target
        y_pred: (np.array), predictions, indices of the buckets where the IV should be computed

    Returns:
        iv (float): information value

    """
    return _IV_score(y_test, y_pred, digits=3)



import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from skorecard.bucketers import DecisionTreeBucketer
from skorecard.reporting import build_bucket_table


def test_report_decision_tree(df):
    """Test the reporting module."""
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"]
    tbt = DecisionTreeBucketer(max_n_bins=4, min_bin_size=0.1, variables=["LIMIT_BAL", "BILL_AMT1"])
    tbt.fit(X, y)
    tbt.transform(X)

    df_out = build_bucket_table(X, y, column="LIMIT_BAL", bucketer=tbt)
    assert df_out.shape == (5, 11)
    # Make sure bucket table equals feature bucket mapping dict
    assert (
        dict(zip(df_out["bucket_id"].values, df_out["label"].values))
        == tbt.features_bucket_mapping_.get("LIMIT_BAL").labels
    )

    expected = pd.DataFrame(
        {"bucket_id": {0: -1, 1: 0, 2: 1, 3: 2, 4: 3}, "Count": {0: 0.0, 1: 849, 2: 676, 3: 1551, 4: 2924}}
    )
    pd.testing.assert_frame_equal(df_out[["bucket_id", "Count"]], expected)

    np.testing.assert_array_equal(
        df_out.columns.ravel(),
        np.array(
            [
                "bucket_id",
                "label",
                "Count",
                "Count (%)",
                "Non-event",
                "Event",
                "% Event",
                "% Non-event",
                "Event Rate",
                "WoE",
                "IV",
            ]
        ),
    )


def test_report_consinstency(df):
    """
    Test that the reported defaults match the ones in the sample.
    """
    X = df[["LIMIT_BAL", "BILL_AMT1"]]
    y = df["default"]
    tbt = DecisionTreeBucketer(max_n_bins=4, min_bin_size=0.1, variables=["LIMIT_BAL", "BILL_AMT1"])
    tbt.fit(X, y)
    tbt.transform(X)

    df_out = build_bucket_table(X, y, column="LIMIT_BAL", bucketer=tbt)
    assert df_out["Event"].sum() == y.sum()

    # adds random selection in the indexing
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25, random_state=42)

    df_out = build_bucket_table(X_train, y_train, column="LIMIT_BAL", bucketer=tbt)
    assert df_out["Event"].sum() == y_train.sum()

    df_out = build_bucket_table(X_test, y_test, column="LIMIT_BAL", bucketer=tbt)
    assert df_out["Event"].sum() == y_test.sum()



import warnings
from typing import Dict, Optional

import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_is_fitted

from skorecard.bucket_mapping import BucketMapping
from skorecard.metrics.metrics import _IV_score


def build_bucket_table(
    X: pd.DataFrame,
    y: np.ndarray,
    column: str,
    bucketer=None,
    bucket_mapping: Optional[BucketMapping] = None,
    epsilon=1,
    display_missing=True,
    verbose=False,
) -> pd.DataFrame:
    """
    Calculates summary statistics for a bucket generated by a skorecard bucketing object.

    This report currently works for just 1 column at a time.

    ``python
    from skorecard import datasets
    from skorecard.bucketers import DecisionTreeBucketer
    from skorecard.reporting import create_report
    X, y = datasets.load_uci_credit_card(return_X_y=True)

    # make sure that those cases
    specials = {
        "LIMIT_BAL":{
            "=50000":[50000],
            "in [20001,30000]":[20000,30000],
            }
    }

    dt_bucketer = DecisionTreeBucketer(variables=['LIMIT_BAL'], specials = specials)
    dt_bucketer.fit(X, y)
    dt_bucketer.transform(X)

    df_report = create_report(X,y,column="LIMIT_BAL", bucketer= dt_bucketer)
    df_report
    ```

    Args:
         X (pd.DataFrame): features
         y (np.array): target
         column (str): column for which you want the report
         bucketer: (optional) Skorecard.bucketing bucketer object. Ignored if bucket_mapping is specified.
         bucket_mapping: (optional) Skorecard.bucket_mapping BucketMapping object
         epsilon(float): small value to prevent zero division error for WoE
         display_missing (boolean): Add a row for missing even when not present in data
         verbose(boolean): be verbose

    Returns:
        df (pandas DataFrame): reporting df
    """
    assert column in X.columns

    X = X.copy()

    if bucket_mapping and bucketer:
        warnings.warn("Both bucket_mapping and bucketer specified. Ignoring bucketer.")

    if not bucket_mapping and not bucketer:
        raise Exception("Specify either bucket_mapping or bucketer")
        # TODO: In case no bucket_mapping and no bucketer specified,use values as-is

    if bucket_mapping and not bucketer:
        col_bucket_mapping = bucket_mapping

    if not bucket_mapping and bucketer:
        bucket_dict = bucketer.features_bucket_mapping_

        col_bucket_mapping = bucket_dict.get(column)

    X_transform = pd.DataFrame(data={"bucket_id": col_bucket_mapping.transform(X[column])}, index=X.index)

    if y is not None:
        X_transform["Event"] = y
    else:
        X_transform["Event"] = np.nan

    # If missing_treatment == passthrough, we reformat the bucket_id and un-do this later
    X_transform["bucket_id"].fillna(31415926535, inplace=True)

    stats = X_transform.groupby("bucket_id", as_index=False).agg(
        def_rate=pd.NamedAgg(column="Event", aggfunc="mean"),
        Event=pd.NamedAgg(column="Event", aggfunc="sum"),
        Count=pd.NamedAgg(column="bucket_id", aggfunc="count"),
    )

    stats["label"] = stats["bucket_id"].map(col_bucket_mapping.labels)
    # Make sure missing is present even when not present
    if display_missing:
        ref = pd.DataFrame.from_dict(col_bucket_mapping.labels, orient="index", columns=["label"])
        ref["bucket_id"] = ref.index
        stats = (
            stats.merge(ref, how="outer", on=["bucket_id", "label"])
            .fillna(0)
            .sort_values("bucket_id")
            .reset_index(drop=True)
        )

    stats["Count (%)"] = np.round(100 * stats["Count"] / stats["Count"].sum(), 2)

    # If unsupervised bucketer, we don't always have y info.
    if y is None:
        columns = ["bucket_id", "label", "Count", "Count (%)"]
        return stats.sort_values(by="bucket_id")[columns]

    stats["Non-event"] = stats["Count"] - stats["Event"]
    # Default rates
    stats["Event Rate"] = stats["Event"] / stats["Count"]  # TODO: can we divide by 0 accidentally?

    stats["% Event"] = stats["Event"] / stats["Event"].sum()
    stats["% Non-event"] = stats["Non-event"] / stats["Non-event"].sum()

    event_percentage = (stats["Event"] + epsilon) / (stats["Event"].sum() + 2 * epsilon)
    non_event_percentage = (stats["Non-event"] + epsilon) / (stats["Non-event"].sum() + 2 * epsilon)
    stats["WoE"] = (event_percentage / non_event_percentage).apply(lambda x: np.log(x))
    stats.loc[stats["Count"] == 0, "WoE"] = np.nan

    stats["IV"] = abs((stats["% Non-event"] - stats["% Event"]) * stats["WoE"])

    stats["% Event"] = np.round(100 * stats["% Event"], 2)
    stats["% Non-event"] = np.round(100 * stats["% Non-event"], 2)
    stats["Event Rate"] = np.round(stats["Event Rate"], 3)

    stats["WoE"] = np.round(stats["WoE"], 3)
    stats["IV"] = np.round(stats["IV"], 3)

    if verbose:
        iv_total = stats["IV"].sum()
        print(f"IV for {column} = {np.round(iv_total, 4)}")

    columns = [
        "bucket_id",
        "label",
        "Count",
        "Count (%)",
        "Non-event",
        "Event",
        "% Event",
        "% Non-event",
        "Event Rate",
        "WoE",
        "IV",
    ]

    # A little reformatting for if missing_treatment is passthrough
    if 31415926535 in stats["bucket_id"].values:
        stats["label"] = np.where(stats["bucket_id"] == 31415926535, "Missing", stats["label"])
        stats = (
            stats.drop_duplicates(subset=["label"], keep="last").reset_index(drop=True).replace([31415926535], np.nan)
        )
    return stats.sort_values(by="bucket_id")[columns]


class BucketTableMethod:
    """
    Add method for bucketing tables to another class.

    To be used with skorecard.pipeline.BucketingProcess and skorecard.bucketers.BaseBucketer
    """

    def bucket_table(self, column):
        """
        Generates the statistics for the buckets of a particular column.

        The pre-buckets are matched to the post-buckets, so that the user has a much clearer understanding of how
        the BucketingProcess ends up with the final buckets.
        An example:

        bucket | label        | Count | Count (%) | Non-event | Event | % Event | % Non-event | Event Rate | WoE  | IV
        -------|--------------|-------|-----------|-----------|-------|---------|-------------|------------|------|-----
        0      | (-inf, 25.0) | 61.0  |	1.36      |	57.0	  | 4.0	  |  0.41	| 1.62        |	0.066      |1.380 |0.017
        1      | [25.0, 45.0) | 2024.0|	44.98	  | 1536.0	  | 488.0 |	49.64	|43.67        |	0.241	   |-0.128|0.008

        Args:
            column: The column we wish to analyse

        Returns:
            df (pd.DataFrame): A pandas dataframe of the format above
        """  # noqa
        if isinstance(self, Pipeline):
            check_is_fitted(self.steps[0][1])
        else:
            check_is_fitted(self)

        if not hasattr(self, "bucket_tables_"):
            raise NotFittedError("You need to fit the bucketer on data in order to calculate the bucket table.")

        if column not in self.bucket_tables_.keys():
            raise ValueError(f"'{column}' is not one of the variables transformed by this bucketer.")

        table = self.bucket_tables_.get(column)
        table = table.rename(columns={"bucket_id": "bucket"})

        return table


class SummaryMethod:
    """
    Adds a `.summary()` method to a bucketing class.

    To be used with skorecard.pipeline.BucketingProcess and skorecard.bucketers.BaseBucketer
    """

    def _generate_summary(self, X, y):
        """
        Calculate the summary table.
        """
        if isinstance(self, Pipeline):
            check_is_fitted(self.steps[-1][1])
        else:
            check_is_fitted(self)

        self.summary_dict_ = {}

        # Calculate information value
        if y is not None:
            iv_scores = iv(self.transform(X), y)
        else:
            iv_scores = {}

        for col in X.columns:
            # In case the column was never (pre)-bucketed
            try:
                prebucket_number = str(len(self.prebucket_tables_[col]["bucket_id"].unique()))
            except KeyError:
                # This specific column was not prebucketed
                prebucket_number = "not_prebucketed"
            except AttributeError:
                # This bucketer did not do any prebucketing
                prebucket_number = "not_prebucketed"

            try:
                bucket_number = str(len(self.bucket_tables_[col]["bucket_id"].unique()))
            except KeyError:
                bucket_number = "not_bucketed"

            self.summary_dict_[col] = [
                prebucket_number,
                bucket_number,
                iv_scores.get(col, "not available"),
                X[col].dtype,
            ]

    def summary(self):
        """
        Display a summary table for columns passed to `.fit()`.

        The format is the following:

        column    | num_prebuckets | num_buckets | dtype
        ----------|----------------|-------------|-------
        LIMIT_BAL |      15        |     10      | float64
        BILL_AMT1 |      15        |     6       | float64
        """  # noqa
        if isinstance(self, Pipeline):
            check_is_fitted(self.steps[-1][1])
        else:
            check_is_fitted(self)

        if not hasattr(self, "summary_dict_"):
            raise NotFittedError("You need to fit Bucketer on data in order to calculate the summary table.")

        return (
            pd.DataFrame.from_dict(
                self.summary_dict_, orient="index", columns=["num_prebuckets", "num_buckets", "IV_score", "dtype"]
            )
            .rename_axis("column")
            .reset_index()
        )


def psi(X1: pd.DataFrame, X2: pd.DataFrame, epsilon=0.0001, digits=None) -> Dict:
    """
    Calculate the PSI between the features in two dataframes, `X1` and `X2`.

    `X1` and `X2` should be bucketed (outputs of fitted bucketers).

    $$
    PSI = \\sum((\\%{ Good } - \\%{ Bad }) \times \\ln \frac{\\%{ Good }}{\\%{ Bad }})
    $$

    Args:
        X1 (pd.DataFrame): bucketed features, expected
        X2 (pd.DataFrame): bucketed features, actual data
        epsilon (float): Amount to be added to relative counts in order to avoid division by zero in the WOE
            calculation.
        digits: (int): number of significant decimal digits in the IV calculation

    Returns: dictionary of psi values. keys are feature names, values are the psi values

    Examples:

    ```python
    from skorecard import datasets
    from sklearn.model_selection import train_test_split
    from skorecard.bucketers import DecisionTreeBucketer
    from skorecard.reporting import psi

    X, y = datasets.load_uci_credit_card(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(
        X,y,
        test_size=0.25,
        random_state=42
    )

    dbt = DecisionTreeBucketer()
    X_train_bins = dbt.fit_transform(X_train,y_train)
    X_test_bins = dbt.transform(X_test)

    psi_dict = psi(X_train_bins, X_test_bins)
    ```
    """  # noqa
    assert (X1.columns == X2.columns).all(), "X1 and X2 must have same columns"

    y1 = pd.Series(0, index=X1.index)
    y2 = pd.Series(1, index=X2.index)

    X = pd.concat([X1, X2], axis=0)
    y = pd.concat([y1, y2], axis=0).reset_index(drop=True)

    psis = {col: _IV_score(y, X[col], epsilon=epsilon, digits=digits) for col in X1.columns}

    return psis


def iv(X: pd.DataFrame, y: pd.Series, epsilon: float = 0.0001, digits: Optional[int] = None) -> Dict:
    r"""
    Calculate the Information Value (IV) of the features in `X`.

    `X` must be the output of fitted bucketers.

    $$
    IV = \sum { (\% goods - \% bads) } * { WOE }
    $$

    $$
    WOE=\ln (\% { goods } /  \% { bads })
    $$

    Example:

    ```python
    from skorecard import datasets
    from sklearn.model_selection import train_test_split
    from skorecard.bucketers import DecisionTreeBucketer
    from skorecard.reporting import iv

    X, y = datasets.load_uci_credit_card(return_X_y=True)

    dbt = DecisionTreeBucketer()
    X_bins = dbt.fit_transform(X,y)

    iv_dict = iv(X_bins, y)
    ```

    Args:
        X: pd.DataFrame (bucketed) features
        y: pd.Series: target values
        epsilon (float): Amount to be added to relative counts in order to avoid division by zero in the WOE
            calculation.
        digits (int): number of significant decimal digits in the IV calculation

    Returns:
        IVs (dict): Keys are feature names, values are the IV values
    """  # noqa
    return {col: _IV_score(y, X[col], epsilon=epsilon, digits=digits) for col in X.columns}



from .pipeline import (
    get_features_bucket_mapping,
    KeepPandas,
    find_bucketing_step,
    SkorecardPipeline,
    to_skorecard_pipeline,
)

from .bucketing_process import BucketingProcess

__all__ = [
    "get_features_bucket_mapping",
    "KeepPandas",
    "find_bucketing_step",
    "BucketingProcess",
    "SkorecardPipeline",
    "to_skorecard_pipeline",
]



from typing import List

from sklearn.base import BaseEstimator, TransformerMixin

from skorecard.utils.validation import ensure_dataframe


class ColumnSelector(BaseEstimator, TransformerMixin):
    """
    Transformer that performs selection of variables from a pandas dataframe.

    Useful in pipelines, where we require a step that selects features.

    Example:

    ```python
    from skorecard import datasets
    from skorecard.preprocessing import ColumnSelector

    X, y = datasets.load_uci_credit_card(return_X_y=True)
    cs = ColumnSelector(variables=['EDUCATION'])
    assert cs.fit_transform(X, y).columns == ['EDUCATION']
    ```
    """

    def __init__(self, variables: List = []):
        """Transformer constructor.

        Args:
            variables: list of columns to select. Default value is set to None - in this case, there is no selection of
                columns.
        """
        self.variables = variables

    def fit(self, X, y=None):
        """
        Fit the transformer.

        Here to be compliant with the sklearn API, does not fit anything.
        """
        # scikit-learn requires checking that X has same shape on transform
        # this is because scikit-learn is still positional based (no column names used)
        self.n_train_features_ = X.shape[1]

        return self

    def transform(self, X):
        """
        Selects the columns.

        Args:
            X (pd.DataFrame): Dataset
        """
        X = ensure_dataframe(X)
        if hasattr(self, "n_train_features_"):
            if X.shape[1] != self.n_train_features_:
                msg = f"Number of features in X ({X.shape[1]}) is different "
                msg += f"from the number of features in X during fit ({self.n_train_features_})"
                raise ValueError(msg)

        if len(self.variables) > 0:
            return X[self.variables]
        else:
            return X

    def _more_tags(self):
        """
        Estimator tags are annotations of estimators that allow programmatic inspection of their capabilities.

        See https://scikit-learn.org/stable/developers/develop.html#estimator-tags
        """  # noqa
        return {"requires_fit": False}



from sklearn.pipeline import make_pipeline

from skorecard.bucketers import DecisionTreeBucketer, OptimalBucketer


def test_full_pipeline(df):
    """Tests some complete pipelines."""
    X = df.drop(columns=["default"])
    y = df["default"]

    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    prebucket_pipeline = make_pipeline(DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05))

    bucket_pipeline = make_pipeline(
        OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
        OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
    )

    pipe = make_pipeline(prebucket_pipeline, bucket_pipeline)

    pipe.fit(X, y)
    pipe.transform(X)



import itertools
import pathlib
from typing import Dict, List, Optional, TypeVar

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.multiclass import unique_labels
from sklearn.utils.validation import check_array, check_is_fitted

from skorecard.bucket_mapping import BucketMapping
from skorecard.features_bucket_mapping import FeaturesBucketMapping
from skorecard.reporting import build_bucket_table
from skorecard.reporting.plotting import PlotBucketMethod
from skorecard.reporting.report import BucketTableMethod, SummaryMethod
from skorecard.utils.exceptions import NotInstalledError
from skorecard.utils.validation import is_fitted

# JupyterDash
try:
    from jupyter_dash import JupyterDash
except ModuleNotFoundError:
    JupyterDash = NotInstalledError("jupyter-dash", "dashboard")

try:
    import dash_bootstrap_components as dbc
except ModuleNotFoundError:
    dbc = NotInstalledError("dash_bootstrap_components", "dashboard")


from skorecard.apps.app_callbacks import add_bucketing_callbacks
from skorecard.apps.app_layout import add_basic_layout
from skorecard.utils.validation import ensure_dataframe

PathLike = TypeVar("PathLike", str, pathlib.Path)


class BaseBucketer(BaseEstimator, TransformerMixin, PlotBucketMethod, BucketTableMethod, SummaryMethod):
    """Base class for bucket transformers."""

    @staticmethod
    def _check_y(y):
        # checks that y is an appropriate type and shape
        if y is None:
            return y

        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise AttributeError("If passing y as a DataFrame, it must be 1 column.")
            y = y.values.reshape(-1)

        if isinstance(y, pd.core.series.Series):
            y = y.values

        if isinstance(y, np.ndarray):
            if y.ndim > 2:
                raise AttributeError("If passing y as a Numpy array, y must be a 1-dimensional")
            elif y.ndim == 2:
                if y.shape[1] != 1:
                    raise AttributeError("If passing y as a Numpy array, y must be a 1-dimensional")
                else:
                    y = y.reshape(-1, 1)

        y = check_array(y, ensure_2d=False)

        return y

    @staticmethod
    def _is_allowed_missing_treatment(missing_treatment):
        # checks if the argument for missing_values is valid
        allowed_str_missing = [
            "separate",
            "most_frequent",
            "most_risky",
            "least_risky",
            "neutral",
            "similar",
            "passthrough",
        ]

        if isinstance(missing_treatment, str):
            if missing_treatment not in allowed_str_missing:
                raise ValueError(f"missing_treatment must be in {allowed_str_missing} or a dict")

        elif isinstance(missing_treatment, dict):
            for _, v in enumerate(missing_treatment):
                if missing_treatment[v] < 0:
                    raise ValueError("As an integer, missing_treatment must be greater than 0")
                elif not isinstance(missing_treatment[v], int):
                    raise ValueError("Values of the missing_treatment dict must be integers")

        else:
            raise ValueError(f"missing_treatment must be in {allowed_str_missing} or a dict")

    @staticmethod
    def _check_contains_na(X, variables: Optional[List]):
        has_missings = X[variables].isnull().any()
        vars_missing = has_missings[has_missings].index.tolist()

        if vars_missing:
            raise ValueError(f"The variables {vars_missing} contain missing values. Consider using an imputer first.")

    @staticmethod
    def _check_variables(X, variables: Optional[List]):
        assert isinstance(variables, list)
        if len(variables) == 0:
            variables = list(X.columns)
        else:
            for var in variables:
                assert var in list(X.columns), f"Column {var} not present in X"
        assert variables is not None and len(variables) > 0
        return variables

    @staticmethod
    def _filter_specials_for_fit(X, y, specials: Dict):
        """
        We need to filter out the specials from a vector.

        Because we don't want to use those values to determine bin boundaries.
        """
        flt_vals = list(itertools.chain(*specials.values()))
        flt = X.isin(flt_vals)
        X_out = X[~flt]

        if y is not None:
            y_out = y[~flt]
        else:
            y_out = y
        return X_out, y_out

    def _find_missing_bucket(self, feature):
        """
        Used for when missing_treatment is in:
        ["most_frequent", "most_risky", "least_risky", "neutral", "similar", "passthrough"]

        Calculates the new bucket for us to put the missing values in.
        """
        if self.missing_treatment == "most_frequent":
            most_frequent_row = (
                self.bucket_tables_[feature].sort_values("Count", ascending=False).reset_index(drop=True).iloc[0]
            )
            if most_frequent_row["label"] != "Missing":
                missing_bucket = int(most_frequent_row["bucket_id"])
            else:
                # missings are already the most common bucket, pick the next one
                missing_bucket = int(
                    self.bucket_tables_[feature]
                    .sort_values("Count", ascending=False)
                    .reset_index(drop=True)["bucket_id"][1]
                )
        elif self.missing_treatment in ["most_risky", "least_risky"]:
            if self.missing_treatment == "least_risky":
                ascending = True
            else:
                ascending = False
            # if fitted with .fit(X) and not .fit(X, y)
            if "Event" not in self.bucket_tables_[feature].columns:
                raise AttributeError("bucketer must be fit with y to determine the risk rates")

            missing_bucket = int(
                self.bucket_tables_[feature][self.bucket_tables_[feature]["bucket_id"] >= 0]
                .sort_values("Event Rate", ascending=ascending)
                .reset_index(drop=True)
                .iloc[0]["bucket_id"]
            )

        elif self.missing_treatment in ["neutral"]:
            table = self.bucket_tables_[feature]
            table["WoE"] = np.abs(table["WoE"])
            missing_bucket = int(
                table[table["Count"] > 0].sort_values("WoE").reset_index(drop=True).iloc[0]["bucket_id"]
            )

        elif self.missing_treatment in ["similar"]:
            table = self.bucket_tables_[feature]
            missing_WoE = table[table["label"] == "Missing"]["WoE"].values[0]
            table["New_WoE"] = np.abs(table["WoE"] - missing_WoE)
            missing_bucket = int(
                table[table["label"] != "Missing"].sort_values("New_WoE").reset_index(drop=True).iloc[0]["bucket_id"]
            )

        elif self.missing_treatment in ["passthrough"]:
            missing_bucket = np.nan

        else:
            raise AssertionError(f"Invalid missing treatment '{self.missing_treatment}' specified")

        return missing_bucket

    def _filter_na_for_fit(self, X: pd.DataFrame, y):
        """
        We need to filter out the missing values from a vector.

        Because we don't want to use those values to determine bin boundaries.

        Note pd.DataFrame.isna and pd.DataFrame.isnull are identical
        """
        # let's also treat infinite values as NA
        # scikit-learn's check_estimator might throw those at us
        with pd.option_context("mode.use_inf_as_na", True):
            flt = pd.isna(X).values
        X_out = X[~flt]
        if y is not None and len(y) > 0:
            y_out = y[~flt]
        else:
            y_out = y
        return X_out, y_out

    @staticmethod
    def _verify_specials_variables(specials: Dict, variables: List) -> None:
        """
        Make sure all specials columns are also in the data.
        """
        diff = set(specials.keys()).difference(set(variables))
        if len(diff) > 0:
            raise ValueError(f"Features {diff} are defined in the specials dictionary, but not in the variables.")

    def fit(self, X, y=None):
        """Fit X, y."""
        # Do __init__ validation.
        # See https://scikit-learn.org/stable/developers/develop.html#parameters-and-init
        assert self.remainder in ["passthrough", "drop"]
        if hasattr(self, "missing_treatment"):
            self._is_allowed_missing_treatment(self.missing_treatment)
        if hasattr(self, "n_bins"):
            assert isinstance(self.n_bins, int)
            assert self.n_bins >= 1
        if hasattr(self, "encoding_method"):
            assert self.encoding_method in ["frequency", "ordered"]
        if hasattr(self, "tol"):
            if self.tol < 0 or self.tol > 1:
                raise ValueError("tol takes values between 0 and 1")
        if hasattr(self, "max_n_categories"):
            if self.max_n_categories is not None:
                if self.max_n_categories < 0 or not isinstance(self.max_n_categories, int):
                    raise ValueError("max_n_categories takes only positive integer numbers")

        # Do data validation
        X = ensure_dataframe(X)
        y = self._check_y(y)
        if y is not None:
            assert len(y) == X.shape[0], "y and X not same length"
            # Store the classes seen during fit
            self.classes_ = unique_labels(y)

        # scikit-learn requires checking that X has same shape on transform
        # this is because scikit-learn is still positional based (no column names used)
        self.n_train_features_ = X.shape[1]

        self.variables_ = self._check_variables(X, self.variables)
        self._verify_specials_variables(self.specials, X.columns)
        self.features_bucket_mapping_ = FeaturesBucketMapping()
        self.bucket_tables_ = {}

        for feature in self.variables_:
            # filter specials for the fit
            if feature in self.specials.keys():
                special = self.specials[feature]
                X_flt, y_flt = self._filter_specials_for_fit(X=X[feature], y=y, specials=special)
            else:
                X_flt, y_flt = X[feature], y
                special = {}

            # filter missing values for the fit
            X_flt, y_flt = self._filter_na_for_fit(X=X_flt, y=y_flt)

            # Find the splits
            # This method is implemented by each bucketer
            assert isinstance(X_flt, pd.Series)
            splits, right = self._get_feature_splits(feature, X=X_flt, y=y_flt, X_unfiltered=X)

            self._update_column_fit(X, y, feature, special, splits, right)

        if self.get_statistics:
            self._generate_summary(X, y)

        return self

    def _update_column_fit(self, X, y, feature, special, splits, right, generate_summary=False):
        """
        Extract out part of the fit for a column.

        Useful when we want to interactively update the fit.
        """
        # Deal with missing values
        missing_bucket = None
        if isinstance(self.missing_treatment, dict):
            missing_bucket = self.missing_treatment.get(feature)

        self.features_bucket_mapping_[feature] = BucketMapping(
            feature_name=feature,
            type=self.variables_type,
            missing_bucket=missing_bucket,
            map=splits,
            right=right,
            specials=special,
        )

        # Calculate the bucket table
        if self.get_statistics:
            self.bucket_tables_[feature] = build_bucket_table(
                X,
                y,
                column=feature,
                bucket_mapping=self.features_bucket_mapping_.get(feature),
            )

            if self.missing_treatment in [
                "most_frequent",
                "most_risky",
                "least_risky",
                "neutral",
                "similar",
                "passthrough",
            ]:
                missing_bucket = self._find_missing_bucket(feature=feature)
                # Repeat above procedure now we know the bucket distribution
                self.features_bucket_mapping_[feature] = BucketMapping(
                    feature_name=feature,
                    type=self.variables_type,
                    missing_bucket=missing_bucket,
                    map=splits,
                    right=right,
                    specials=special,
                )

                # Recalculate the bucket table with the new bucket for missings
                self.bucket_tables_[feature] = build_bucket_table(
                    X,
                    y,
                    column=feature,
                    bucket_mapping=self.features_bucket_mapping_.get(feature),
                )

        if generate_summary:
            self._generate_summary(X, y)

    def fit_interactive(self, X, y=None, mode="external", **server_kwargs):
        """
        Fit a bucketer and then interactive edit the fit using a dash app.

        Note we are using a [jupyterdash](https://medium.com/plotly/introducing-jupyterdash-811f1f57c02e) app,
        which supports 3 different modes:

        - 'external' (default): Start dash server and print URL
        - 'inline': Start dash app inside an Iframe in the jupyter notebook
        - 'jupyterlab': Start dash app as a new tab inside jupyterlab

        """
        # We need to make sure we only fit if not already fitted
        # This prevents a user losing manually defined boundaries
        # when re-running .fit_interactive()
        if not is_fitted(self):
            self.fit(X, y)

        self.app = JupyterDash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
        add_basic_layout(self)
        add_bucketing_callbacks(self, X, y)
        self.app.run_server(mode=mode, **server_kwargs)

    def transform(self, X: pd.DataFrame, y=None) -> pd.DataFrame:
        """Transforms an array into the corresponding buckets fitted by the Transformer.

        Args:
            X (pd.DataFrame): dataframe which will be transformed into the corresponding buckets
            y (array): target

        Returns:
            df (pd.DataFrame): dataset with transformed features
        """
        check_is_fitted(self)
        X = ensure_dataframe(X)
        y = self._check_y(y)
        if y is not None:
            assert len(y) == X.shape[0], "y and X not same length"
        # If bucketer was fitted, make sure X has same columns as train
        if hasattr(self, "n_train_features_"):
            if X.shape[1] != self.n_train_features_:
                raise ValueError("number of features in transform is different from the number of features in fit")

        # Some bucketers do not have a .fit() method
        # and if user did not specify any variables
        # use all the variables defined in the features_bucket_mapping
        if not hasattr(self, "variables_"):
            if self.variables == []:
                self.variables_ = list(self.features_bucket_mapping_.maps.keys())
            else:
                self.variables_ = self.variables

        for feature in self.variables_:
            bucket_mapping = self.features_bucket_mapping_.get(feature)
            X[feature] = bucket_mapping.transform(X[feature])

        if self.remainder == "drop":
            return X[self.variables_]
        else:
            return X

    def predict(self, X: pd.DataFrame):
        """Applies the transform method. To be used for the grid searches.

        Args:
            X (pd.DataFrame): The numerical data which will be transformed into the corresponding buckets

        Returns:
            y (np.array): Transformed X, such that the values of X are replaced by the corresponding bucket numbers
        """
        return self.transform(X)

    def predict_proba(self, X: pd.DataFrame):
        """Applies the transform method. To be used for the grid searches.

        Args:
            X (pd.DataFrame): The numerical data which will be transformed into the corresponding buckets

        Returns:
            yhat (np.array): transformed X, such that the values of X are replaced by the corresponding bucket numbers
        """
        return self.transform(X)

    def save_yml(self, fout: PathLike) -> None:
        """
        Save the features bucket to a yaml file.

        Args:
            fout: file output
        """
        check_is_fitted(self)
        if isinstance(self.features_bucket_mapping_, dict):
            FeaturesBucketMapping(self.features_bucket_mapping_).save_yml(fout)
        else:
            self.features_bucket_mapping_.save_yml(fout)

    def _more_tags(self):
        """
        Estimator tags are annotations of estimators that allow programmatic inspection of their capabilities.

        See https://scikit-learn.org/stable/developers/develop.html#estimator-tags
        """  # noqa
        return {"binary_only": True, "allow_nan": True}



import numpy as np
import pandas as pd
import pytest
from sklearn.compose import ColumnTransformer
from sklearn.exceptions import NotFittedError
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler

from skorecard.bucket_mapping import BucketMapping
from skorecard.bucketers import (
    DecisionTreeBucketer,
    EqualFrequencyBucketer,
    EqualWidthBucketer,
    OptimalBucketer,
    OrdinalCategoricalBucketer,
)
from skorecard.pipeline import BucketingProcess, KeepPandas, find_bucketing_step, get_features_bucket_mapping
from skorecard.pipeline.pipeline import SkorecardPipeline, to_skorecard_pipeline
from skorecard.utils import BucketingPipelineError


@pytest.mark.filterwarnings("ignore:sklearn.")
def test_keep_pandas(df, caplog):
    """Tests the KeepPandas() class."""
    y = df["default"].values
    X = df.drop(columns=["default", "pet_ownership"])

    bucket_pipeline = make_pipeline(
        StandardScaler(),
        EqualWidthBucketer(n_bins=5, variables=["LIMIT_BAL", "BILL_AMT1"]),
    )
    # Doesn't work, input should be a pandas dataframe.
    with pytest.raises(AssertionError):
        bucket_pipeline.fit(X, y)

    bucket_pipeline = make_pipeline(
        KeepPandas(StandardScaler()),
        EqualWidthBucketer(n_bins=5, variables=["LIMIT_BAL", "BILL_AMT1"]),
    )

    with pytest.raises(NotFittedError):
        bucket_pipeline.transform(X)

    bucket_pipeline.fit(X, y)
    assert isinstance(bucket_pipeline.transform(X), pd.DataFrame)

    bucket_pipeline = ColumnTransformer(
        [
            ("categorical_preprocessing", OrdinalCategoricalBucketer(), ["EDUCATION", "MARRIAGE"]),
            ("numerical_preprocessing", EqualWidthBucketer(n_bins=5), ["LIMIT_BAL", "BILL_AMT1"]),
        ],
        remainder="passthrough",
    )

    # Make sure warning is raised
    caplog.clear()
    KeepPandas(make_pipeline(bucket_pipeline))
    assert "sklearn.compose.ColumnTransformer can change" in caplog.text

    # Make sure warning is raised
    caplog.clear()
    KeepPandas(bucket_pipeline)
    assert "sklearn.compose.ColumnTransformer can change" in caplog.text

    assert isinstance(KeepPandas(bucket_pipeline).fit_transform(X, y), pd.DataFrame)


def test_bucketing_pipeline(df):
    """Test the class."""
    y = df["default"].values
    X = df.drop(columns=["default"])

    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    prebucket_pipeline = make_pipeline(DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05))

    bucket_pipeline = make_pipeline(
        OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
        OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
    )

    pipe = make_pipeline(prebucket_pipeline, bucket_pipeline)
    pipe.fit(X, y)
    # Make sure we can fit it twice
    pipe.fit(X, y)

    # make sure transforms work.
    pipe.transform(X)
    pipe.fit_transform(X, y)


def test_find_coarse_classing_step(df):
    """Tests coarse classing step."""
    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    bucketing_process = BucketingProcess(
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05)
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )
    pipeline = make_pipeline(bucketing_process)
    assert find_bucketing_step(pipeline, identifier="bucketingprocess") == 0


def test_get_features_bucket_mapping(df):
    """Test retrieving info from sklearn pipeline."""
    y = df["default"].values
    X = df.drop(columns=["default"])

    nested_pipeline = make_pipeline(
        make_pipeline(EqualWidthBucketer(n_bins=5, variables=["LIMIT_BAL", "BILL_AMT1"])),
        OrdinalCategoricalBucketer(variables=["EDUCATION", "MARRIAGE"]),
    )

    with pytest.raises(NotFittedError):
        get_features_bucket_mapping(nested_pipeline)

    nested_pipeline.fit(X, y)
    bm = get_features_bucket_mapping(nested_pipeline)
    assert (
        bm.get("EDUCATION").map
        == BucketMapping(feature_name="EDUCATION", type="categorical", map={2: 0, 1: 1, 3: 2}, right=True).map
    )


# TODO: write tests with different kinds of sklearn pipelines
# - ColumnTransformer and ColumnSelector usage


def test_make_pipeline(df):
    """Make sure bucketers work inside a pipeline."""
    y = df["default"].values
    X = df.drop(columns=["default"])

    pipe = make_pipeline(
        EqualWidthBucketer(n_bins=4, variables=["LIMIT_BAL"]),
        EqualFrequencyBucketer(n_bins=7, variables=["BILL_AMT1"]),
    )
    new_X = pipe.fit_transform(X, y)
    assert isinstance(new_X, pd.DataFrame)


def test_pipeline_errors(df):
    """Make sure incorrect input also throws correct errors in pipeline."""
    y = df["default"].values
    X = df.drop(columns=["default"])

    bu = EqualWidthBucketer(n_bins=4, variables=["LIMIT_BAL", "BILL_AMT1"])
    with pytest.raises(NotFittedError):
        bu.transform(X)  # not fitted yet
    with pytest.raises(ValueError):
        bu.fit_transform(np.array([1, 2, 3]), y)


def test_pipeline_has_no_duplicated_features(df):
    """Test that the columns in bucketers in a scikit pipeline are unique, otherwise an error is raised."""
    y = df["default"].values
    X = df.drop(columns=["default"])
    features_1 = ["LIMIT_BAL", "BILL_AMT1", "EDUCATION"]
    features_2 = ["EDUCATION", "MARRIAGE", "BILL_AMT1"]

    bucketer = make_pipeline(
        DecisionTreeBucketer(variables=features_1, max_n_bins=5),
        OrdinalCategoricalBucketer(variables=features_2, tol=0.05),
    )

    # during fit
    with pytest.raises(BucketingPipelineError):
        to_skorecard_pipeline(bucketer).fit(X, y)

    # after fit
    bucketer.fit(X, y)
    with pytest.raises(BucketingPipelineError):
        to_skorecard_pipeline(bucketer)

    with pytest.raises(BucketingPipelineError):
        SkorecardPipeline(
            [
                ("dtb", DecisionTreeBucketer(variables=features_1, max_n_bins=5)),
                ("ocb", OrdinalCategoricalBucketer(variables=features_2, tol=0.05)),
            ]
        ).fit(X, y)

    # What if one of the pipelines applies to all
    # And the other to some?
    # That should also be a conflict
    bucketer = make_pipeline(
        DecisionTreeBucketer(max_n_bins=5),
        OrdinalCategoricalBucketer(variables=features_2, tol=0.05),
    )
    with pytest.raises(BucketingPipelineError):
        p = to_skorecard_pipeline(bucketer)
        feat = ["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "MARRIAGE"]
        p.fit(X[feat], y)


def test_skorecard_pipeline(df):
    """Test that the skorecard Pipeline returns the same results."""
    features = ["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "MARRIAGE"]
    X = df[features]
    y = df["default"].values

    features_1 = ["LIMIT_BAL", "BILL_AMT1"]
    features_2 = ["EDUCATION", "MARRIAGE"]

    bucketer = make_pipeline(
        DecisionTreeBucketer(variables=features_1, max_n_bins=5),
        OrdinalCategoricalBucketer(variables=features_2, tol=0.05),
    )

    bucketer.fit(X, y)

    sk_pipe = to_skorecard_pipeline(bucketer)

    assert bucketer.transform(X).equals(sk_pipe.transform(X))
    assert isinstance(sk_pipe.summary(), pd.DataFrame)
    assert isinstance(sk_pipe.bucket_table("LIMIT_BAL"), pd.DataFrame)
    assert isinstance(sk_pipe.bucket_table("EDUCATION"), pd.DataFrame)

    # Now with specials
    bucketer = make_pipeline(
        DecisionTreeBucketer(
            variables=features_1, max_n_bins=5, specials={"BILL_AMT1": {"Some specials": [201800, 76445, 79244]}}
        ),
        OrdinalCategoricalBucketer(variables=features_2, tol=0.05, specials={"EDUCATION": {"Some specials": [1, 2]}}),
    )
    bucketer.fit(X, y)
    sk_pipe = to_skorecard_pipeline(bucketer)


# TODO Cleanup
# def test_bucket_transformer_bin_count_list(df):
#     """Test the exception is raised in scikit-learn pipeline."""
#     with pytest.raises(AttributeError):
#         transformer = ColumnTransformer(
#             transformers=[
#                 ("simple", SimpleBucketTransformer(bin_count=2), [1]),
#                 ("agglom", AgglomerativeBucketTransformer(bin_count=4), [0]),
#                 ("quantile", QuantileBucketTransformer(bin_count=[10]), [3]),
#             ],
#             remainder="passthrough",
#         )
#         transformer.fit_transform(df.values)

#     return None


# def test_bucket_transformer_exception(df):
#     """Test the exception is raised in scikit-learn pipeline."""
#     with pytest.raises(DimensionalityError):
#         transformer = ColumnTransformer(
#             transformers=[
#                 ("simple", SimpleBucketTransformer(bin_count=2), [1]),
#                 ("agglom", AgglomerativeBucketTransformer(bin_count=4), [0]),
#                 ("quantile", QuantileBucketTransformer(bin_count=10), [2, 3]),
#             ],
#             remainder="passthrough",
#         )
#         transformer.fit_transform(df.values)

#     return None


# def test_bucket_transformer(df):
#     """Test that we can utilise the main bucket transformers in a scikit-learn pipeline."""
#     transformer = ColumnTransformer(
#         transformers=[
#             ("simple", SimpleBucketTransformer(bin_count=2), [1]),
#             ("agglom", AgglomerativeBucketTransformer(bin_count=4), [0]),
#             ("quantile_0", QuantileBucketTransformer(bin_count=10), [2]),
#             ("quantile_1", QuantileBucketTransformer(bin_count=6), [3]),
#         ],
#         remainder="passthrough",
#     )

#     X = transformer.fit_transform(df.values)

#     # Test only non-categorical variables
#     assert len(np.unique(X[:, 2])) == 10
#     assert len(np.unique(X[:, 3])) == 6
#     assert np.all(X[:, 4] == df["default"].values)

#     return None



import numpy as np
import numpy.testing as npt
import pandas as pd
import pytest
from sklearn.pipeline import make_pipeline

from skorecard.bucketers import AsIsCategoricalBucketer, DecisionTreeBucketer, OptimalBucketer
from skorecard.pipeline import BucketingProcess
from skorecard.preprocessing import WoeEncoder

# TODO: WoE should treat missing values as a separate bin and thus handled seamlessly.


@pytest.fixture()
def X_y():
    """Set of X,y for testing the transformers."""
    X = pd.DataFrame(
        np.array(
            [[0, 1], [1, 0], [0, 0], [3, 2], [0, 1], [1, 2], [2, 0], [2, 1], [0, 0]],
            np.int32,
        ),
        columns=["col1", "col2"],
    )
    y = pd.Series(np.array([0, 0, 0, 1, 1, 1, 0, 0, 1]))

    return X, y


@pytest.fixture()
def X_y_2():
    """Set of X,y for testing the transformers.

    In the first column, bucket 3 is not present in class 1.
    """
    X = pd.DataFrame(
        np.array(
            [[0, 1], [1, 0], [0, 0], [3, 2], [0, 1], [1, 2], [2, 0], [2, 1], [0, 0]],
            np.int32,
        ),
        columns=["col1", "col2"],
    )
    y = pd.Series(np.array([0, 0, 0, 0, 0, 1, 1, 1, 1]))

    return X, y


def test_woe_transformed_dimensions(X_y):
    """Tests that the total number of unique WOEs matches the unique number of bins in X."""
    X, y = X_y

    woeb = WoeEncoder(variables=["col1", "col2"])
    new_X = woeb.fit_transform(X, y)
    assert len(new_X["col1"].unique()) == len(X["col1"].unique())
    assert len(new_X["col2"].unique()) == len(X["col2"].unique())


def test_missing_bucket(X_y_2):
    """Tests that the total number of unique WOEs matches the unique number of bins in X."""
    X, y = X_y_2

    woeb = WoeEncoder(variables=["col1", "col2"])
    new_X = woeb.fit_transform(X, y)

    assert new_X.shape == X.shape
    assert len(new_X["col1"].unique()) == len(X["col1"].unique())
    assert len(new_X["col2"].unique()) == len(X["col2"].unique())

    # because class 1 will have zero counts, the WOE transformer will divide by the value of epsilon, avoiding infinite
    # numbers
    assert not any(new_X["col1"] == np.inf)
    assert not any(new_X["col2"] == np.inf)

    # If epsilon is set to zero, expect a Division By Zero exception
    with pytest.raises(ZeroDivisionError):
        WoeEncoder(epsilon=0, variables=["col1", "col2"]).fit_transform(X, y)


def test_woe_values(X_y):
    """Tests the value of the WOE."""
    X, y = X_y

    woeb = WoeEncoder(variables=["col1", "col2"])
    new_X = woeb.fit_transform(X, y)
    new_X

    expected = pd.DataFrame(
        {
            "col1": {
                0: -0.22309356256166865,
                1: -0.22304359629388562,
                2: -0.22309356256166865,
                3: -7.824445930877619,
                4: -0.22309356256166865,
                5: -0.22304359629388562,
                6: 8.294299608857235,
                7: 8.294299608857235,
                8: -0.22309356256166865,
            },
            "col2": {
                0: 0.469853677979616,
                1: 0.8752354701118937,
                2: 0.8752354701118937,
                3: -8.517393171418904,
                4: 0.469853677979616,
                5: -8.517393171418904,
                6: 0.8752354701118937,
                7: 0.469853677979616,
                8: 0.8752354701118937,
            },
        }
    )

    pd.testing.assert_frame_equal(new_X, expected)

    assert woeb.transform(X).equals(new_X)


def test_woe_in_pipeline(df):
    """Tests if WoeEncoder works in pipeline context."""
    y = df["default"]
    X = df.drop(columns=["default"])

    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    bucketing_process = BucketingProcess(
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            AsIsCategoricalBucketer(variables=cat_cols),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )

    pipeline = make_pipeline(
        bucketing_process,
        WoeEncoder(),
    )

    out = pipeline.fit_transform(X, y)
    assert out.select_dtypes(include=[float]).shape[1] == out.shape[1]
    npt.assert_almost_equal(out["LIMIT_BAL"][0], 0.566607, decimal=4)
    npt.assert_almost_equal(out["LIMIT_BAL"][1], -0.165262, decimal=4)
    npt.assert_almost_equal(out["LIMIT_BAL"][3], -0.031407, decimal=4)
    npt.assert_almost_equal(out["BILL_AMT1"][0], 0.124757, decimal=4)
    npt.assert_almost_equal(out["BILL_AMT1"][1], 0.091788, decimal=4)
    npt.assert_almost_equal(out["BILL_AMT1"][3], -0.103650, decimal=4)



import pytest
from sklearn.linear_model import LogisticRegression
from sklearn.pipeline import make_pipeline

from skorecard.bucketers import DecisionTreeBucketer, OptimalBucketer
from skorecard.bucketers.bucketers import AsIsCategoricalBucketer, UserInputBucketer
from skorecard.pipeline import BucketingProcess
from skorecard.pipeline.bucketing_process import _find_remapped_specials
from skorecard.preprocessing import WoeEncoder
from skorecard.utils import NotBucketedError, NotBucketObjectError, NotPreBucketedError


def test_bucketing_process_order(df):
    """Test that a NotPreBucketedError is raised if the bucketing pipeline is passed before the prebucketing."""
    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]
    X = df[num_cols + cat_cols]
    y = df["default"].values

    # The prebucketing pipeline does not process all numerical columns in bucketing pipeline
    bucketing_process = BucketingProcess(
        specials={"LIMIT_BAL": {"=400000.0": [400000.0]}},
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=["LIMIT_BAL"], max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )
    with pytest.raises(NotBucketedError):
        bucketing_process.fit(X, y)

    # The bucketing pipeline does not process all numerical columns in prebucketing pipeline
    bucketing_process = BucketingProcess(
        specials={"LIMIT_BAL": {"=400000.0": [400000.0]}},
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=["LIMIT_BAL"], max_n_bins=100, min_bin_size=0.05),
            AsIsCategoricalBucketer(variables=cat_cols),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )
    with pytest.raises(NotPreBucketedError):
        bucketing_process.fit(X, y)


def test_non_bucketer_in_pipeline(df):
    """Test that putting a non-bucketer in bucket_process raises error."""
    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]
    X = df[num_cols + cat_cols]
    y = df["default"].values

    with pytest.raises(NotBucketObjectError):
        # input validation is only done on fit, as per scikitlearn convention
        BucketingProcess(
            prebucketing_pipeline=make_pipeline(
                DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05)
            ),
            bucketing_pipeline=make_pipeline(
                OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
                OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
                LogisticRegression(),  # Should break the process
            ),
        ).fit(X, y)

    with pytest.raises(NotBucketObjectError):
        BucketingProcess(
            prebucketing_pipeline=make_pipeline(
                DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
                LogisticRegression(),  # Should break the process
            ),
            bucketing_pipeline=make_pipeline(
                OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
                OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
            ),
        ).fit(X, y)


def test_bucketing_optimization(df):
    """Test that the optimal bucketer returns less or equal number of unique buckets."""
    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    X = df[num_cols + cat_cols]
    y = df["default"].values

    bucketing_process = BucketingProcess(
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            AsIsCategoricalBucketer(variables=cat_cols),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )

    X_bins = bucketing_process.fit_transform(X, y)

    X_prebucketed = bucketing_process.pre_pipeline_.transform(X)
    for col in num_cols + cat_cols:
        assert X_bins[col].nunique() <= X_prebucketed[col].nunique()
        assert X_bins[col].nunique() > 1


def test_bucketing_with_specials(df):
    """Test that specials propagate."""
    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    X = df[num_cols + cat_cols]
    y = df["default"].values

    the_specials = {"LIMIT_BAL": {"=400000.0": [400000.0]}}

    bucketing_process = BucketingProcess(
        specials=the_specials,
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            AsIsCategoricalBucketer(variables=cat_cols),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )

    _ = bucketing_process.fit_transform(X, y)

    # Make sure all the prebucketers have the specials assigned
    for step in bucketing_process.pre_pipeline_:
        assert step.specials == the_specials

    # Test the specials in the prebucket table
    prebuckets = bucketing_process.prebucket_table("LIMIT_BAL")
    assert prebuckets["Count"][0] == 45.0
    assert prebuckets["label"][0] == "Special: =400000.0"

    # Test the specials in the bucket table
    buckets = bucketing_process.bucket_table("LIMIT_BAL")
    assert buckets["Count"][0] == 45.0
    assert buckets["label"][0] == "Special: =400000.0"


def test_bucketing_process_in_pipeline(df):
    """Test that it works fine within a sklearn pipeline."""
    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    X = df[num_cols + cat_cols]
    y = df["default"].values

    bucketing_process = BucketingProcess(
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            AsIsCategoricalBucketer(variables=cat_cols),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )

    pipeline = make_pipeline(bucketing_process, WoeEncoder(), LogisticRegression())

    pipeline.fit(X, y)
    preds = pipeline.predict_proba(X)

    assert preds.shape[0] == X.shape[0]


def test_bucketing_process_with_numerical_specials(df):
    """
    Test we get expected results for numerical specials.
    """
    y = df["default"]
    X = df.drop(columns=["default"])

    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    bucketing_process = BucketingProcess(
        specials={"LIMIT_BAL": {"=400000.0": [400000.0]}},
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            AsIsCategoricalBucketer(variables=cat_cols),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )
    bucketing_process.fit(X, y)

    table = bucketing_process.prebucket_table("LIMIT_BAL")
    assert len(table["bucket"].unique()) == 11
    assert table[["label"]].values[0] == "Special: =400000.0"

    table = bucketing_process.prebucket_table("MARRIAGE")
    assert table.shape[0] == 6


def test_bucketing_process_with_categorical_specials(df):
    """
    Test we get expected results for numerical specials.
    """
    y = df["default"]
    X = df.drop(columns=["default"])

    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    bucketing_process = BucketingProcess(
        specials={"MARRIAGE": {"=0": [0]}},
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            DecisionTreeBucketer(variables=cat_cols, max_n_bins=100, min_bin_size=0.05),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )
    bucketing_process.fit(X, y)

    table = bucketing_process.prebucket_table("MARRIAGE")
    assert table.shape[0] == 4
    assert table["label"][0] == "Special: =0"


def test_uib_process(df):
    """
    Test merging works properly.
    """
    y = df["default"]
    X = df.drop(columns=["default"])

    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    bucketing_process = BucketingProcess(
        specials={"MARRIAGE": {"=0": [0]}, "LIMIT_BAL": {"=400000.0": [400000.0]}},
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            AsIsCategoricalBucketer(variables=cat_cols),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )

    bucketing_process.fit(X, y)

    uib = UserInputBucketer(bucketing_process.features_bucket_mapping_)

    X_trans = bucketing_process.transform(X)
    assert X_trans.equals(uib.transform(X))


def test_bucketing_process_summary(df):
    """
    Test bucketing process.

    Test we get expected results for .summary()
    """
    y = df["default"]
    X = df.drop(columns=["default"])

    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    bucketing_process = BucketingProcess(
        specials={"MARRIAGE": {"=0": [0]}},
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            DecisionTreeBucketer(
                variables=cat_cols, max_n_bins=100, min_bin_size=0.05
            ),  # note this is wrong.. it's now treated as a numerical feature
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )

    bucketing_process.fit(X, y)
    table = bucketing_process.summary()
    assert set(table.columns) == {"column", "num_prebuckets", "num_buckets", "IV_score", "dtype"}
    assert table[table["column"] == "pet_ownership"]["num_prebuckets"].values[0] == "not_prebucketed"
    assert table[table["column"] == "pet_ownership"]["num_buckets"].values[0] == "not_bucketed"
    assert len(table["dtype"].unique()) == 3
    assert all(table["IV_score"] >= 0)


def test_bucketing_process_remainder(df):
    """
    Test bucketing process.

    Test we get expected results for .summary()
    """
    y = df["default"]
    X = df.drop(columns=["default", "pet_ownership"])

    num_cols = ["LIMIT_BAL"]  # left out "BILL_AMT1"
    cat_cols = ["EDUCATION"]  # left out "MARRIAGE"

    # Don't drop variables.
    # If we specify bucketingprocess.variables
    # other variables should not be altered by prebucketing or bucketing pipeline
    bucketing_process = BucketingProcess(
        variables=num_cols + cat_cols,
        remainder="passthrough",
        specials={"EDUCATION": {"=0": [0]}},
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(max_n_bins=100, min_bin_size=0.05),  # note we didn't specify variables here!
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(max_n_bins=10, min_bin_size=0.05),  # note we didn't specify variables here!
        ),
    )
    new_X = bucketing_process.fit_transform(X, y)
    assert new_X["MARRIAGE"].equals(X["MARRIAGE"])
    assert len(new_X.columns) == 4

    # now with remainder drop
    bucketing_process = BucketingProcess(
        variables=num_cols + cat_cols,
        remainder="drop",
        specials={"EDUCATION": {"=0": [0]}},
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            DecisionTreeBucketer(
                variables=cat_cols, max_n_bins=100, min_bin_size=0.05
            ),  # note it's now treated as a numerical feature
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.05),
        ),
    )
    new_X = bucketing_process.fit_transform(X, y)
    assert len(new_X.columns) == 2


def test_remapping_specials():
    """
    Test remapping works.
    """
    bucket_labels = {
        0: "(-inf, 25000.0)",
        1: "[25000.0, 45000.0)",
        2: "[45000.0, 55000.0)",
        3: "[55000.0, 75000.0)",
        4: "[75000.0, 85000.0)",
        5: "[85000.0, 105000.0)",
        6: "[105000.0, 145000.0)",
        7: "[145000.0, 175000.0)",
        8: "[175000.0, 225000.0)",
        9: "[225000.0, 275000.0)",
        10: "[275000.0, 325000.0)",
        11: "[325000.0, 385000.0)",
        12: "[385000.0, inf)",
        13: "Missing",
        14: "Special: =400000.0",
    }

    var_specials = {"=400000.0": [400000.0]}

    assert _find_remapped_specials(bucket_labels, var_specials) == {"=400000.0": [14]}

    assert _find_remapped_specials(bucket_labels, None) == {}
    assert _find_remapped_specials(None, None) == {}

    bucket_labels = {13: "Special: =12345 or 123456", 14: "Special: =400000.0"}

    var_specials = {"=400000.0": [400000.0], "=12345 or 123456": [12345, 123456]}
    assert _find_remapped_specials(bucket_labels, var_specials) == {"=400000.0": [14], "=12345 or 123456": [13]}



from skorecard.apps.app_utils import colorize_cell, perc_data_bars
from skorecard.utils.exceptions import NotInstalledError

# Dash + dependencies
try:
    try:
        from dash import dash_table, dcc, html
    except ImportError:
        import dash_core_components as dcc
        import dash_html_components as html
        import dash_table
except ModuleNotFoundError:
    dcc = NotInstalledError("dash_core_components", "dashboard")
    html = NotInstalledError("dash_html_components", "dashboard")
    Input = NotInstalledError("dash", "dashboard")
    Output = NotInstalledError("dash", "dashboard")
    State = NotInstalledError("dash", "dashboard")
    dash_table = NotInstalledError("dash_table", "dashboard")

# Dash Bootstrap
try:
    import dash_bootstrap_components as dbc
except ModuleNotFoundError:
    dbc = NotInstalledError("dash_bootstrap_components", "dashboard")


# the style arguments for the sidebar.
SIDEBAR_STYLE = {
    "position": "fixed",
    "top": 0,
    "left": 0,
    "bottom": 0,
    "width": "20%",
    "padding": "20px 10px",
    "background-color": "#f8f9fa",
}
# the style arguments for the main content page.
CONTENT_STYLE = {"margin-left": "25%", "margin-right": "5%", "top": 0, "padding": "20px 10px"}
TEXT_STYLE = {"textAlign": "center", "color": "#191970"}
CARD_TEXT_STYLE = {"textAlign": "center", "color": "#0074D9"}


def add_basic_layout(self):
    """
    Adds a basic layout to self.app.
    """
    column_options = [{"label": o, "value": o} for o in self.features_bucket_mapping_.columns]

    sidebar = get_sidebar(self, column_options)

    content = html.Div(
        children=[
            dbc.Row(
                [
                    html.H3(
                        [
                            html.Span("Current column", id="column_title"),
                            dbc.Badge("tbd", id="column_type", className="ml-1"),
                        ],
                        style=TEXT_STYLE,
                    ),
                ]
            ),
            html.Br(),
            dbc.Row(
                [
                    html.H4("Bucket plot", style=TEXT_STYLE),
                ]
            ),
            dbc.Row([dcc.Graph(id="graph-bucket", style={"width": "100%"})]),
            dbc.Row(
                [
                    html.H4("Bucket table", style=TEXT_STYLE),
                ]
            ),
            html.Br(),
            dbc.Row(
                [
                    dash_table.DataTable(
                        id="bucket_table",
                        style_table={
                            "width": "100%",
                            "minWidth": "100%",
                        },
                        style_data={"whiteSpace": "normal", "height": "auto"},
                        style_cell={
                            "height": "auto",
                            "overflow": "hidden",
                            "textOverflow": "ellipsis",
                            "minWidth": "80px",
                            "textAlign": "left",
                        },
                        style_cell_conditional=[
                            {"if": {"column_id": "range"}, "width": "180px"},
                        ],
                        page_size=20,
                        columns=[
                            {"name": "label", "id": "label", "editable": False},
                            {"name": "bucket", "id": "bucket", "editable": False},
                            {"name": "Count", "id": "Count", "editable": False},
                            {"name": "Count (%)", "id": "Count (%)", "editable": False},
                            {"name": "Non-event", "id": "Non-event", "editable": False},
                            {"name": "Event", "id": "Event", "editable": False},
                            {"name": "Event Rate", "id": "Event Rate", "editable": False},
                            {"name": "WoE", "id": "WoE", "editable": False},
                            {"name": "IV", "id": "IV", "editable": False},
                        ],
                        style_data_conditional=perc_data_bars("Count (%)")
                        + perc_data_bars("Event Rate")
                        + [
                            {"if": {"row_index": "odd"}, "backgroundColor": "rgb(248, 248, 248)"},
                            {
                                "if": {"column_editable": True},
                                "backgroundColor": "rgb(46,139,87)",
                                "color": "white",
                                "border": "1px solid black",
                            },
                            {
                                "if": {"state": "active"},
                                "backgroundColor": "rgba(0, 116, 217, 0.3)",
                                "border": "1px solid rgb(0, 116, 217)",
                            },
                        ]
                        + colorize_cell("bucket"),
                        style_header_conditional=[
                            {
                                "if": {"column_editable": True},
                                "border": "1px solid black",
                            }
                        ],
                        style_header={
                            "backgroundColor": "rgb(230, 230, 230)",
                        },
                        editable=False,
                        fill_width=True,
                    ),
                ],
                style={"width": "100%"},
            ),
        ],
        style=CONTENT_STYLE,
    )

    self.app.layout = html.Div([sidebar, content])


def add_bucketing_process_layout(self):
    """
    Adds a layout to self.app.
    """
    column_options = [{"label": o, "value": o} for o in self.features_bucket_mapping_.columns]
    sidebar = get_sidebar(self, column_options)

    content = html.Div(
        children=[
            dbc.Row(
                [
                    html.H3(
                        [
                            html.Span("Current column", id="column_title"),
                            dbc.Badge("tbd", id="column_type", className="ml-1"),
                        ],
                        style=TEXT_STYLE,
                    ),
                ]
            ),
            html.Br(),
            dbc.Row(
                [
                    html.H4("Pre-bucket plot", style=TEXT_STYLE),
                ]
            ),
            dbc.Row([dcc.Graph(id="graph-prebucket", style={"width": "100%"})]),
            dbc.Row(
                [
                    html.H4("Pre-bucket table", style=TEXT_STYLE),
                ]
            ),
            html.Br(),
            dbc.Row(
                [
                    dbc.Badge(
                        "There is something wrong with your bucketing info",
                        color="danger",
                        id="bucket-error-msg",
                        style={"display": "none"},
                    ),
                    dash_table.DataTable(
                        id="pre_bucket_table",
                        style_table={
                            "width": "100%",
                            "minWidth": "100%",
                        },
                        style_data={"whiteSpace": "normal", "height": "auto"},
                        style_cell={
                            "height": "auto",
                            "overflow": "hidden",
                            "textOverflow": "ellipsis",
                            "minWidth": "80px",
                            "textAlign": "left",
                        },
                        style_cell_conditional=[
                            {"if": {"column_id": "range"}, "width": "180px"},
                        ],
                        page_size=20,
                        columns=[
                            {"name": "label", "id": "label", "editable": False},
                            {"name": "pre-bucket", "id": "pre-bucket", "editable": False},
                            {"name": "Count", "id": "Count", "editable": False},
                            {"name": "Count (%)", "id": "Count (%)", "editable": False},
                            {"name": "Non-event", "id": "Non-event", "editable": False},
                            {"name": "Event", "id": "Event", "editable": False},
                            {"name": "Event Rate", "id": "Event Rate", "editable": False},
                            {"name": "WoE", "id": "WoE", "editable": False},
                            {"name": "IV", "id": "IV", "editable": False},
                            {"name": "bucket", "id": "bucket", "editable": True},
                        ],
                        style_data_conditional=perc_data_bars("Count (%)")
                        + perc_data_bars("Event Rate")
                        + [
                            {"if": {"row_index": "odd"}, "backgroundColor": "rgb(248, 248, 248)"},
                            {
                                "if": {"column_editable": True},
                                "backgroundColor": "rgb(46,139,87)",
                                "color": "white",
                                "border": "1px solid black",
                            },
                            {
                                "if": {"state": "active"},
                                "backgroundColor": "rgba(0, 116, 217, 0.3)",
                                "border": "1px solid rgb(0, 116, 217)",
                            },
                        ]
                        + colorize_cell("bucket"),
                        style_header_conditional=[
                            {
                                "if": {"column_editable": True},
                                "border": "1px solid black",
                            }
                        ],
                        style_header={
                            "backgroundColor": "rgb(230, 230, 230)",
                        },
                        editable=False,
                        fill_width=True,
                    ),
                ],
                style={"width": "100%"},
            ),
            html.Br(),
            dbc.Row(
                [
                    html.H4("Bucket table", style=TEXT_STYLE),
                ]
            ),
            html.Br(),
            dbc.Row(
                [
                    dash_table.DataTable(
                        id="bucket_table",
                        style_table={
                            "width": "100%",
                            "minWidth": "100%",
                        },
                        style_data={"whiteSpace": "normal", "height": "auto"},
                        style_cell={
                            "height": "auto",
                            "overflow": "hidden",
                            "textOverflow": "ellipsis",
                            "minWidth": "80px",
                            "textAlign": "left",
                        },
                        style_cell_conditional=[
                            {"if": {"column_id": "range"}, "width": "180px"},
                        ],
                        page_size=20,
                        columns=[
                            {"name": "bucket", "id": "bucket", "editable": False},
                            {"name": "label", "id": "label", "editable": False},
                            {"name": "Count", "id": "Count", "editable": False},
                            {"name": "Count (%)", "id": "Count (%)", "editable": False},
                            {"name": "Non-event", "id": "Non-event", "editable": False},
                            {"name": "Event", "id": "Event", "editable": False},
                            {"name": "Event Rate", "id": "Event Rate", "editable": False},
                            {"name": "WoE", "id": "WoE", "editable": False},
                            {"name": "IV", "id": "IV", "editable": False},
                        ],
                        style_data_conditional=perc_data_bars("Count (%)")
                        + perc_data_bars("Event Rate")
                        + [
                            {"if": {"row_index": "odd"}, "backgroundColor": "rgb(248, 248, 248)"},
                            {
                                "if": {"column_editable": True},
                                "backgroundColor": "rgb(46,139,87)",
                                "color": "white",
                                "border": "1px solid black",
                            },
                            {
                                "if": {"state": "active"},
                                "backgroundColor": "rgba(0, 116, 217, 0.3)",
                                "border": "1px solid rgb(0, 116, 217)",
                            },
                        ]
                        + colorize_cell("bucket"),
                        style_header_conditional=[
                            {
                                "if": {"column_editable": True},
                                "border": "1px solid black",
                            }
                        ],
                        style_header={
                            "backgroundColor": "rgb(230, 230, 230)",
                        },
                        editable=False,
                        fill_width=True,
                    ),
                ],
                style={"width": "100%"},
            ),
            html.Br(),
            dbc.Row(
                [
                    html.H4("Bucket plot", style=TEXT_STYLE),
                ]
            ),
            dbc.Row([dcc.Graph(id="graph-bucket", style={"width": "100%"})]),
        ],
        style=CONTENT_STYLE,
    )

    self.app.layout = html.Div([sidebar, content])


def get_sidebar(self, column_options):
    """
    Layout for the sidebar in the dash app.
    """
    return html.Div(
        [
            html.H5(f"{type(self).__name__}", style=TEXT_STYLE),
            html.Hr(),
            dbc.Label("Select feature", html_for="input_column"),
            dcc.Dropdown(
                id="input_column",
                options=column_options,
                value=column_options[0].get("label"),
            ),
            html.Br(),
            dbc.FormGroup(
                [
                    dbc.Label("Map", html_for="input_map"),
                    dbc.Textarea(
                        id="input_map",
                        style={"minHeight": 160},
                    ),
                    dbc.FormText(
                        "You can manually set different boundaries", color="secondary", id="input_map_helptext"
                    ),
                    dbc.FormFeedback("Feedback will be given here", valid=False, id="input_map_feedback"),
                ]
            ),
            dbc.FormGroup(
                [
                    dbc.Label("Bucket assignment"),
                    dbc.FormText(
                        "You can edit bucket assignment in the highlighted right column of the prebucket table.",
                    ),
                ],
                id="bucketingprocess-helptext",
                style={"display": "none"},
            ),
            html.H5("Export"),
            html.P(
                [
                    "Your updated buckets are saved to the class instance, ",
                    "which means you can safely close this app. See also ",
                    dcc.Link(
                        "working with manually defined buckets",
                        href="https://ing-bank.github.io/skorecard/howto/using_manually_defined_buckets/",
                    ),
                ],
                className="small",
            ),
            html.P(
                [
                    "For convenience, you can also copy this generated code snippet defining a ",
                    dcc.Link(
                        html.Code("UserInputBucketer"),
                        href="https://ing-bank.github.io/skorecard/api/bucketers/UserInputBucketer/",
                    ),
                    dcc.Clipboard(content="hi", title="code snippet", style={"font-size": "120%"}, id="code_export"),
                ],
                className="small",
            ),
        ],
        style=SIDEBAR_STYLE,
    )



import numpy as np
import pandas as pd
import pytest

from skorecard.bucketers import OrdinalCategoricalBucketer


def test_threshold_min(df) -> None:
    """Test that threshold_min < 1 raises an error."""
    with pytest.raises(ValueError):
        OrdinalCategoricalBucketer(tol=-0.1, variables=["EDUCATION"]).fit(X=None, y=None)
    with pytest.raises(ValueError):
        OrdinalCategoricalBucketer(tol=1.001, variables=["EDUCATION"]).fit(X=None, y=None)


@pytest.mark.parametrize("tol_input, expected_unique,", [(0.44, 2), (0.05, 4), (0, 7), (0.5, 1)])
def test_correct_output(df, tol_input, expected_unique):
    """Test that correct use of CatBucketTransformer returns expected results."""
    X = df
    y = df["default"].values

    cbt = OrdinalCategoricalBucketer(tol=tol_input, variables=["EDUCATION"])
    cbt.fit(X, y)
    X_trans = cbt.transform(X)
    assert len(X_trans["EDUCATION"].unique()) == expected_unique


def test_mapping_dict(df):
    """Test that the mapping dict is created correctly."""
    X = df
    y = df["default"].values
    cbt = OrdinalCategoricalBucketer(tol=0, variables=["EDUCATION"])
    cbt.fit(X, y)
    bucket_map = cbt.features_bucket_mapping_.get("EDUCATION")
    assert len(bucket_map.map) == len(np.unique(X["EDUCATION"]))


def test_encoding_method(df):
    """Test the encoding method."""
    X = df[["EDUCATION", "default"]]
    y = df["default"]

    ocb = OrdinalCategoricalBucketer(tol=0.03, variables=["EDUCATION"], encoding_method="frequency")
    ocb.fit(X, y)

    assert ocb.features_bucket_mapping_.get("EDUCATION").map == {1: 1, 2: 0, 3: 2}

    ocb = OrdinalCategoricalBucketer(tol=0.03, variables=["EDUCATION"], encoding_method="ordered")
    ocb.fit(X, y)

    assert ocb.features_bucket_mapping_.get("EDUCATION").map == {1: 0, 2: 2, 3: 1}


def test_specials(df):
    """Test specials get assigned to the highest bin."""
    X = df[["EDUCATION"]]
    y = df["default"]

    ocb = OrdinalCategoricalBucketer(
        tol=0.03, variables=["EDUCATION"], encoding_method="ordered", specials={"EDUCATION": {"ed 0": [1]}}
    )
    ocb.fit(X, y)

    X_transform = ocb.transform(X)
    # Make sure value 1 is assigned bucket -3
    assert np.unique(X_transform[X["EDUCATION"] == 1].values)[0] == -3

    ocb = OrdinalCategoricalBucketer(
        tol=0.03, variables=["EDUCATION"], encoding_method="frequency", specials={"EDUCATION": {"ed 0": [1]}}
    )
    ocb.fit(X, y)

    X_transform = ocb.transform(X)
    # Make sure value 1 is assigned bucket -3
    assert np.unique(X_transform[X["EDUCATION"] == 1].values)[0] == -3


def test_missing_default(df_with_missings) -> None:
    """Test that missing values are assigned to the right bucket."""
    X = df_with_missings
    y = df_with_missings["default"].values

    bucketer = OrdinalCategoricalBucketer(variables=["EDUCATION"])
    bucketer = OrdinalCategoricalBucketer(max_n_categories=2, variables=["EDUCATION"])
    X["EDUCATION_trans"] = bucketer.fit_transform(X[["EDUCATION"]], y)

    assert len(X["EDUCATION_trans"].unique()) == 3  # 2 + 1 for NAs
    assert X[X["EDUCATION"].isnull()].shape[0] == X[X["EDUCATION_trans"] == -1].shape[0]


def test_missing_manual(df_with_missings) -> None:
    """Test that missing values are assigned to the right bucket."""
    X = df_with_missings
    y = df_with_missings["default"].values

    bucketer = OrdinalCategoricalBucketer(variables=["EDUCATION"])
    bucketer = OrdinalCategoricalBucketer(
        max_n_categories=2, variables=["EDUCATION"], missing_treatment={"EDUCATION": 0}
    )
    X["EDUCATION_trans"] = bucketer.fit_transform(X[["EDUCATION"]], y)

    assert len(X["EDUCATION_trans"].unique()) == 2
    assert X[X["EDUCATION"].isnull()]["EDUCATION_trans"].sum() == 0


def test_missings():
    """
    Test proper handling of NAs.
    """
    X = pd.DataFrame({"colour": ["blue"] * 10 + ["red"] * 10 + ["grey"] * 10 + [np.nan]})
    y = [1] * 5 + [0] * 5 + [1] * 8 + [0] * 2 + [1] * 1 + [0] * 9 + [1]
    y = np.array(y)
    ocb = OrdinalCategoricalBucketer()
    X_trans = ocb.fit_transform(X, y)
    buckets = X_trans["colour"].unique().tolist()
    buckets.sort()

    assert buckets == [-1, 0, 1, 2]
    assert ocb.bucket_table("colour").shape[0] == 5

    ocb = OrdinalCategoricalBucketer(encoding_method="ordered")
    X_trans = ocb.fit_transform(X, y)
    assert buckets == [-1, 0, 1, 2]
    assert ocb.bucket_table("colour").shape[0] == 5



"""
Tests all the codeblocks in the docstrings, making sure they execute.

This approach is adapted from, and explained in:
https://calmcode.io/docs/epic.html

Try it out with:

```bash
pytest tests/test_docstring.py --verbose
```
"""

import pytest

from tests.conftest import CLASSES_TO_TEST, FUNCTIONS_TO_TEST


def get_public_methods(cls_ref):
    """Helper test function, gets all public methods in a class."""
    return [m for m in dir(cls_ref) if m == "__init__" or not m.startswith("_")]


def get_test_pairs(classes_to_test):
    """Helper test function, get tuples with class and public method."""
    test_pairs = []
    for cls_ref in classes_to_test:
        for meth_ref in get_public_methods(cls_ref):
            test_pairs.append((cls_ref, meth_ref))
    return test_pairs


def handle_docstring(doc, indent):
    """
    Check python code in docstring.

    This function will read through the docstring and grab
    the first python code block. It will try to execute it.
    If it fails, the calling test should raise a flag.
    """
    if not doc:
        return
    start = doc.find("```python\n")
    end = doc.find("```\n")
    if start != -1:
        if end != -1:
            code_part = doc[(start + 10) : end].replace(" " * indent, "")
            print(code_part)
            exec(code_part)


# @pytest.mark.parametrize("m", MODULES_TO_TEST) # TODO Cleanup
# def test_module_docstrings(m):
#     """
#     Take the docstring of a given module.

#     The test passes if the usage examples causes no errors.
#     """
#     handle_docstring(m.__doc__, indent=0)


@pytest.mark.parametrize("c", CLASSES_TO_TEST)
def test_class_docstrings(c):
    """
    Take the docstring of a given class.

    The test passes if the usage examples causes no errors.
    """
    handle_docstring(c.__doc__, indent=4)


@pytest.mark.parametrize("clf_ref,meth_ref", get_test_pairs(CLASSES_TO_TEST))
def test_method_docstrings(clf_ref, meth_ref):
    """
    Take the docstring of every method (m) on the class (c).

    The test passes if the usage examples causes no errors.
    """
    handle_docstring(getattr(clf_ref, meth_ref).__doc__, indent=8)


@pytest.mark.parametrize("f", FUNCTIONS_TO_TEST)
def test_function_docstrings(f):
    """
    Take the docstring of every function.

    The test passes if the usage examples causes no errors.
    """
    handle_docstring(f.__doc__, indent=4)



import os
from contextlib import contextmanager

import pytest
import yaml
from sklearn.pipeline import make_pipeline

from skorecard.bucketers import (
    AgglomerativeClusteringBucketer,
    AsIsCategoricalBucketer,
    AsIsNumericalBucketer,
    DecisionTreeBucketer,
    EqualFrequencyBucketer,
    EqualWidthBucketer,
    OptimalBucketer,
    OrdinalCategoricalBucketer,
    UserInputBucketer,
)
from skorecard.pipeline import BucketingProcess, to_skorecard_pipeline

BUCKETERS_WITH_SET_BINS = [EqualWidthBucketer, AgglomerativeClusteringBucketer, EqualFrequencyBucketer]

AS_IS_BUCKETERS = [AsIsNumericalBucketer, AsIsCategoricalBucketer]


@contextmanager
def working_directory(path):
    """
    Temporary working directories.

    A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.

    Usage:

    ```python
    # Do something in original directory
    with working_directory('/my/new/path'):
        # Do something in new directory
    # Back to old directory
    ````

    Credits: https://gist.github.com/nottrobin/3d675653244f8814838a
    """
    prev_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


@pytest.fixture
def fpath(tmpdir):
    """
    Temporary path.
    """
    return tmpdir.mkdir("sub")


def test_ordinal_bucketer_to_file(df, tmpdir):
    """Test that saving the bucketer to file and loading it back in the user input bucketer works."""
    features = ["EDUCATION", "MARRIAGE"]
    X = df[features]
    y = df["default"]

    ocb = OrdinalCategoricalBucketer(variables=["EDUCATION", "MARRIAGE"], tol=0.1)
    X_trans = ocb.fit_transform(X, y)[features]

    # Test save to yaml in bucketer
    ocb.save_yml(open(os.path.join(tmpdir, "buckets.yml"), "w"))
    buckets_yaml = yaml.safe_load(open(os.path.join(tmpdir, "buckets.yml")))
    X_trans_yaml = UserInputBucketer(buckets_yaml).transform(X)
    assert X_trans.equals(X_trans_yaml)

    # Test save to yaml with str path
    with working_directory(tmpdir):
        ocb.save_yml("buckets.yml")
        buckets_yaml = yaml.safe_load(open("buckets.yml"))
        X_trans_yaml = UserInputBucketer(buckets_yaml).transform(X)
        assert X_trans.equals(X_trans_yaml)

        # test alternative flow
        ocb.save_yml("buckets.yml")
        X_trans_yaml = UserInputBucketer("buckets.yml").transform(X)
        assert X_trans.equals(X_trans_yaml)


def test_decision_tree_bucketer_to_file(df, tmpdir):
    """Test that saving the bucketer to file and loading it back in the user input bucketer works."""
    features = ["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "MARRIAGE"]
    X = df[features]
    y = df["default"]

    tbt = DecisionTreeBucketer(
        variables=features, max_n_bins=5, dt_kwargs={"criterion": "entropy", "min_impurity_decrease": 0.001}
    )
    X_trans = tbt.fit_transform(X, y)

    # Test save to yaml in bucketer
    tbt.save_yml(open(os.path.join(tmpdir, "buckets.yml"), "w"))
    buckets_yaml = yaml.safe_load(open(os.path.join(tmpdir, "buckets.yml")))
    X_trans_yaml = UserInputBucketer(buckets_yaml).transform(X)
    assert X_trans.equals(X_trans_yaml)


@pytest.mark.parametrize("bucketer", BUCKETERS_WITH_SET_BINS)
def test_bucketers_with_n_bins_to_file(bucketer, df, tmpdir) -> None:
    """Test that saving the bucketer to file and loading it back in the user input bucketer works."""
    features = ["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "MARRIAGE"]
    X = df[features]
    y = df["default"].values

    BUCK = bucketer(n_bins=3, variables=features)
    BUCK.fit(X)
    X_trans = BUCK.fit_transform(X, y)

    # Test save to yaml in bucketer
    BUCK.save_yml(open(os.path.join(tmpdir, "buckets.yml"), "w"))
    buckets_yaml = yaml.safe_load(open(os.path.join(tmpdir, "buckets.yml")))
    X_trans_yaml = UserInputBucketer(buckets_yaml).transform(X)
    assert X_trans.equals(X_trans_yaml)


def test_bucketers_with_sklearn_pipeline(df, tmpdir):
    """
    Test bucketers saved in a sklearn pipeline.
    """
    features = ["LIMIT_BAL", "BILL_AMT1", "EDUCATION", "MARRIAGE"]
    X = df[features]
    y = df["default"]

    bucketing = make_pipeline(
        DecisionTreeBucketer(
            variables=[features[0]], max_n_bins=5, dt_kwargs={"criterion": "entropy", "min_impurity_decrease": 0.001}
        ),
        DecisionTreeBucketer(variables=[features[1]], max_n_bins=3),
    )

    X_trans = bucketing.fit_transform(X, y)
    bucketing = to_skorecard_pipeline(bucketing)

    # Test save to yaml
    bucketing.save_yml(open(os.path.join(tmpdir, "buckets.yml"), "w"))
    buckets_yaml = yaml.safe_load(open(os.path.join(tmpdir, "buckets.yml")))
    X_trans_yaml = UserInputBucketer(buckets_yaml).transform(X)
    assert X_trans.equals(X_trans_yaml)


def test_bucketing_process_to_file(df, tmpdir):
    """Test that saving the bucketing_process to file and loading it back in the user input bucketer works."""
    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    X = df[num_cols + cat_cols]
    y = df["default"].values

    bucketing_process = BucketingProcess(
        prebucketing_pipeline=make_pipeline(
            DecisionTreeBucketer(variables=num_cols, max_n_bins=100, min_bin_size=0.05),
            OrdinalCategoricalBucketer(variables=cat_cols, tol=0.01),
        ),
        bucketing_pipeline=make_pipeline(
            OptimalBucketer(variables=num_cols, max_n_bins=10, min_bin_size=0.05),
            OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=10, min_bin_size=0.1),
        ),
    )

    X_trans = bucketing_process.fit_transform(X, y)

    # Test save to yaml in bucketer
    bucketing_process.save_yml(open(os.path.join(tmpdir, "buckets.yml"), "w"))
    buckets_yaml = yaml.safe_load(open(os.path.join(tmpdir, "buckets.yml")))
    assert buckets_yaml == bucketing_process.features_bucket_mapping_.as_dict()
    # Test transforms work the same
    X_trans_yaml = UserInputBucketer(buckets_yaml).transform(X)
    assert X_trans.equals(X_trans_yaml)


@pytest.mark.parametrize("bucketer", AS_IS_BUCKETERS)
def test_as_is_bucketers_to_file(bucketer, df, tmpdir) -> None:
    """Test that saving the bucketer to file and loading it back in the user input bucketer works."""
    features = ["LIMIT_BAL", "EDUCATION", "MARRIAGE"]  # "BILL_AMT1" has >100 unique values, so cannot be AsIs Bucketed
    X = df[features]
    y = df["default"].values

    BUCK = bucketer()
    BUCK.fit(X)
    X_trans = BUCK.fit_transform(X, y)

    # Test save to yaml in bucketer
    BUCK.save_yml(open(os.path.join(tmpdir, "buckets.yml"), "w"))
    buckets_yaml = yaml.safe_load(open(os.path.join(tmpdir, "buckets.yml")))
    X_trans_yaml = UserInputBucketer(buckets_yaml).transform(X)
    assert X_trans.equals(X_trans_yaml)

    # Test inside a skorecard pipeline
    pipe = to_skorecard_pipeline(make_pipeline(BUCK))
    pipe.fit(X, y)
    assert X_trans.equals(pipe.transform(X))
    pipe.save_yml(open(os.path.join(tmpdir, "buckets3.yml"), "w"))
    buckets_yaml3 = yaml.safe_load(open(os.path.join(tmpdir, "buckets3.yml")))
    X_trans_yaml = UserInputBucketer(buckets_yaml3).transform(X)
    assert X_trans.equals(X_trans_yaml)



import warnings

from skorecard import Skorecard
from skorecard.datasets import load_uci_credit_card


def test_suppressor_warning():
    """Checks suppressor effect warning on Skorecard fit."""
    # Load the data. Construct datasets with and without suppressor effect occurring
    data = load_uci_credit_card()
    y = data["target"]

    X_no_suppression = data["data"]
    model = Skorecard()

    X_suppression = X_no_suppression.copy()
    X_suppression["suppressor"] = X_suppression[X_suppression.columns[0]] - X_suppression[X_suppression.columns[1]]
    model_suppression = Skorecard()

    with warnings.catch_warnings(record=True) as w:
        # Check that the suppressor warning is not issued and that no coefficient has an unexpected sign
        model = model.fit(X_no_suppression, y)
        relevant_warning_issued = 0
        if len(w) > 0:
            latest_warning = str(w[-1].message)
            msg = (
                "Features found with coefficient-sign that is contrary to what is expected based on weight-of-evidence."
            )
            relevant_warning_issued = msg in latest_warning
        coefs = model.coef_[0]
        suppression = any(c < 0 for c in coefs)
        assert (not relevant_warning_issued) & (not suppression)

        # Check that the suppressor warning is issued and that there is a coefficient with an unexpected sign
        model_suppression = model_suppression.fit(X_suppression, y)
        relevant_warning_issued = 0
        if len(w) > 0:
            latest_warning = str(w[-1].message)
            msg = (
                "Features found with coefficient-sign that is contrary to what is expected based on weight-of-evidence."
            )
            relevant_warning_issued = msg in latest_warning
        coefs = model_suppression.coef_[0]
        suppression = any(c < 0 for c in coefs)
        assert (relevant_warning_issued) & (suppression)



import dataclasses

import yaml

from skorecard.bucket_mapping import BucketMapping, merge_bucket_mapping


class FeaturesBucketMapping:
    """Stores a collection of features BucketMapping.

    ```python
    from skorecard.bucket_mapping import BucketMapping
    from skorecard.features_bucket_mapping import FeaturesBucketMapping

    # Working with collections of BucketMappings
    bucket1 = BucketMapping(feature_name='feature1', type='numerical', map=[2, 3, 4, 5])
    bucket2 = BucketMapping(feature_name='feature2', type='numerical', map=[5,6,7,8])
    features_bucket_mapping = FeaturesBucketMapping([bucket1, bucket2])
    print(features_bucket_mapping)

    # You can also work with class as dict
    features_bucket_mapping.as_dict()

    features_dict = {
        'feature1': {'feature_name': 'feature1',
            'type': 'numerical',
            'map': [2, 3, 4, 5],
            'right': True},
        'feature2': {'feature_name': 'feature2',
            'type': 'numerical',
            'map': [5, 6, 7, 8],
            'right': True}
    }

    features_bucket_mapping = FeaturesBucketMapping()
    features_bucket_mapping.load_dict(features_dict)
    # Or directly from dict
    FeaturesBucketMapping(features_dict)
    # See columns
    features_bucket_mapping.columns
    ```
    """

    def __init__(self, maps=[]):
        """Takes list of bucketmappings and stores as a dict.

        Args:
            maps (list): list of BucketMapping. Defaults to [].
        """
        self.maps = {}
        if isinstance(maps, list):
            for bucketmap in maps:
                self.append(bucketmap)

        if isinstance(maps, dict):
            for _, bucketmap in maps.items():
                if not isinstance(bucketmap, BucketMapping):
                    bucketmap = BucketMapping(**bucketmap)
                self.append(bucketmap)

    def __repr__(self):
        """Pretty print self.

        Returns:
            str: reproducible object representation.
        """
        class_name = self.__class__.__name__
        maps = list(self.maps.values())
        return f"{class_name}({maps})"

    def __len__(self):
        """
        Length of the map.
        """
        return len(self.maps)

    def __eq__(self, other):
        """
        Define equality.
        """
        return self.maps == other.maps

    def __getitem__(self, key):
        """
        Retrieve BucketMappings by feature name.
        """
        return self.maps[key]

    def __setitem__(self, key, value):
        """
        Set a bucketmapping using the feature name.
        """
        self.maps[key] = value

    def get(self, col: str):
        """Get BucketMapping for a column.

        Args:
            col (str): Name of column

        Returns:
            mapping (BucketMapping): BucketMapping for column
        """
        return self.maps[col]

    def append(self, bucketmap: BucketMapping) -> None:
        """Add a BucketMapping to the collection.

        Args:
            bucketmap (BucketMapping): map of a feature
        """
        assert isinstance(bucketmap, BucketMapping)
        self.maps[bucketmap.feature_name] = bucketmap

    def load_yml(self) -> None:
        """Should load in data from a yml.

        Returns:
            None: nothing
        """
        raise NotImplementedError("todo")

    def save_yml(self, file) -> None:
        """Should write data to a yml.

        Returns:
            None: nothing
        """
        if isinstance(file, str):
            file = open(file, "w")
        yaml.safe_dump(self.as_dict(), file)

    def load_dict(self, obj):
        """Should load in data from a python dict.

        Args:
            obj (dict): Dict with names of features and their BucketMapping

        Returns:
            None: nothing
        """
        assert isinstance(obj, dict)

        self.maps = {}
        for feature, bucketmap in obj.items():
            self.append(BucketMapping(**bucketmap))

    def as_dict(self):
        """Returns data in class as a dict.

        Returns:
            dict: Data in class
        """
        return {k: dataclasses.asdict(v) for k, v in self.maps.items()}

    @property
    def columns(self):
        """Returns the columns that have a bucket_mapping."""
        return list(self.as_dict().keys())


def merge_features_bucket_mapping(a: FeaturesBucketMapping, b: FeaturesBucketMapping) -> FeaturesBucketMapping:
    """
    Merge two sets of sequentual FeatureBucketMapping.

    If there are unique features, we'll add them as-in.
    """
    assert isinstance(a, FeaturesBucketMapping)
    assert isinstance(b, FeaturesBucketMapping)

    cols_in_both = [col for col in a.columns if col in b.columns]
    cols_in_a = [col for col in a.columns if col not in b.columns]
    cols_in_b = [col for col in b.columns if col not in a.columns]

    features_bucket_mapping = FeaturesBucketMapping()

    for col in cols_in_both:
        c = merge_bucket_mapping(a.get(col), b.get(col))
        features_bucket_mapping.append(c)

    for col in cols_in_a:
        features_bucket_mapping.append(a.get(col))

    for col in cols_in_b:
        features_bucket_mapping.append(b.get(col))

    return features_bucket_mapping



import pandas as pd
import pytest

from skorecard.apps.app_utils import determine_boundaries
from skorecard.bucket_mapping import BucketMapping


@pytest.mark.parametrize(
    "map_input, right_input,",
    [([2, 3, 4, 5], True), ([2, 3, 4, 5], False), ([3], True), ([3], False), ([], True), ([], False)],
)
def test_determine_boundaries(map_input, right_input):
    """Tests function."""
    df = pd.DataFrame()
    df["pre_buckets"] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    bucket_mapping = BucketMapping("feature1", "numerical", map=map_input, right=right_input)
    df["buckets"] = bucket_mapping.transform(df["pre_buckets"])
    assert bucket_mapping.map == determine_boundaries(df, bucket_mapping)



import warnings
from typing import List, Optional

import numpy as np
from category_encoders.woe import WOEEncoder
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.utils.validation import check_is_fitted

from skorecard.bucketers import DecisionTreeBucketer, OptimalBucketer
from skorecard.linear_model import LogisticRegression
from skorecard.pipeline import BucketingProcess, to_skorecard_pipeline
from skorecard.pipeline.pipeline import _get_all_steps
from skorecard.preprocessing import ColumnSelector
from skorecard.utils import BucketerTypeError
from skorecard.utils.validation import check_suppressor_effect, ensure_dataframe, is_fitted

ignores = [
    (
        "category_encoders",
        FutureWarning,
        "is_categorical is deprecated and will be removed in a future version.  Use is_categorical_dtype instead",
    )
    # add addition warnings you want to ignore
]

for module, category, message in ignores:
    warnings.filterwarnings("ignore", module=module, category=category, message=message)


class Skorecard(BaseEstimator, ClassifierMixin):
    """Scikit-learn-like estimator that builds a scorecard model.

    Usage Examples:

    usage of Skorecard() without bucketing

    ```python
    from skorecard import Skorecard
    from skorecard.datasets import load_uci_credit_card

    X,y = load_uci_credit_card(return_X_y=True)
    model = Skorecard(variables = ["LIMIT_BAL", "BILL_AMT1","EDUCATION", "MARRIAGE"])
    model.fit(X, y)
    ```

    Alternatively, by passing a predefined Bucketing Process.

    ```python
    from skorecard import Skorecard
    from skorecard.pipeline import BucketingProcess
    from skorecard.bucketers import DecisionTreeBucketer, OrdinalCategoricalBucketer, OptimalBucketer
    from skorecard.datasets import load_uci_credit_card
    from sklearn.pipeline import make_pipeline

    X,y = load_uci_credit_card(return_X_y=True)
    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    prebucketing_pipeline = make_pipeline(
        DecisionTreeBucketer(variables=num_cols, max_n_bins=100),
        OrdinalCategoricalBucketer(variables=cat_cols, tol=0.01)
    )
    bucketing_pipeline = make_pipeline(
        OptimalBucketer(variables=num_cols, max_n_bins=5, min_bin_size=0.08),
        OptimalBucketer(variables=cat_cols, variables_type="categorical", max_n_bins=5, min_bin_size=0.08)
    )

    bucketer = BucketingProcess(prebucketing_pipeline = prebucketing_pipeline, bucketing_pipeline = bucketing_pipeline)
    skorecard_model = Skorecard(bucketing=bucketer, variables=num_cols+cat_cols)
    skorecard_model.fit(X, y)

    # Details
    skorecard_model.bucket_table("LIMIT_BAL")
    skorecard_model.plot_bucket("LIMIT_BAL")
    skorecard_model.prebucket_table("LIMIT_BAL")
    skorecard_model.plot_prebucket("LIMIT_BAL")
    ```

    Using it with bucketers:

     ```python
    from skorecard import Skorecard
    from skorecard.bucketers import DecisionTreeBucketer, OrdinalCategoricalBucketer
    from skorecard.datasets import load_uci_credit_card
    from sklearn.pipeline import make_pipeline

    X,y = load_uci_credit_card(return_X_y=True)
    num_cols = ["LIMIT_BAL", "BILL_AMT1"]
    cat_cols = ["EDUCATION", "MARRIAGE"]

    bucket_pipe  = make_pipeline(
        DecisionTreeBucketer(variables=num_cols, max_n_bins=5),
        OrdinalCategoricalBucketer(variables=cat_cols, tol=0.05)
    )

    skorecard_model = Skorecard(bucketing=bucket_pipe, variables=num_cols+cat_cols)
    skorecard_model.fit(X, y)

    # Details
    skorecard_model.bucket_table("LIMIT_BAL")
    skorecard_model.plot_bucket("LIMIT_BAL")
    ```

    """

    def __init__(
        self,
        bucketing=None,
        *,
        specials: dict = {},
        encoder: str = "woe",
        variables: List = [],
        verbose: int = 0,
        random_state: Optional[int] = None,
        lr_kwargs: dict = {"solver": "lbfgs"},
        calculate_stats: bool = False,
    ):
        """
        Init the class.

        Args:
            bucketing: bucketing step, can be a bucketer, a pipeline of bucketers or a BucketingProcess.
                    Default value is None. In that case it will generate a predefined BucketingProcess where
                    numerical and categorical columns will attempt to be auto-detected and bucketed accordingly.
            specials: (nested) dictionary of special values that require their own binning.
                    Used only if bucketing=None.
                    The dictionary has the following format:
                        {"<column name>" : {"name of special bucket" : <list with 1 or more values>}}
                    For every feature that needs a special value, a dictionary must be passed as value.
                    This dictionary contains a name of a bucket (key) and an array of unique values that should be put
                    in that bucket.
                    When special values are defined, they are not considered in the fitting procedure.
            encoder (string): indicating the type of encoder. Currently only 'woe' (weight-of-evidence) is supported.
            variables (list): list of features after bucketing to fit the LogisticRegression model on. Defaults to None (all features selected).
            verbose (int): verbosity, set to 0 to avoid warning methods.
            random_state (int): the random state that is passed to the LogisticRegression and all Bucketers that have this attribute
            lr_kwargs (dict): Settings passed to skorecard.linear_model.LogisticRegression.
                By default no settings are passed.
            calculate_stats (bool): Passed to skorecard.linear_model.LogisticRegression.
                Will calculate statistic like standard error at the cost of fit() time. Defaults to False.
        """  # noqa
        self.bucketing = bucketing
        self.specials = specials
        self.encoder = encoder
        self.variables = variables
        self.verbose = verbose
        self.random_state = random_state
        self.lr_kwargs = lr_kwargs
        self.calculate_stats = calculate_stats
        if random_state is not None:
            self.lr_kwargs.update({"random_state": self.random_state})

    def __repr__(self):
        """Pretty print self.

        Returns:
            str: reproducible object representation.
        """
        vars = ""
        for k, v in self.__dict__.items():
            if k == "bucketing":
                v = str(v.__class__.__name__)  # get only the class name of the bucketer
            vars += f"{k}={v}, "

        return f"{self.__class__.__name__}({vars})"

    def _build_default_bucketing_process(self, X):
        """
        A default BucketingProcess to use if not specified by user.
        """
        # Auto-detect cat and numerical columns
        num_features = list(X._get_numeric_data().columns)
        cat_features = [f for f in X.columns if f not in num_features]

        # Features with less than 10 unique values should be treated at categoricals
        # Because there is probably no point in binning them further.
        for col in num_features:
            if X[col].nunique() < 10:
                cat_features.append(col)
                num_features.remove(col)

        prebucketing_pipe = []
        bucketing_pipe = []

        if len(num_features) > 0:
            prebucketing_pipe.append(DecisionTreeBucketer(variables=num_features, max_n_bins=100, min_bin_size=0.03))
            bucketing_pipe.append(OptimalBucketer(variables=num_features, max_n_bins=10, min_bin_size=0.05))
        if len(cat_features) > 0:
            bucketing_pipe.append(
                OptimalBucketer(
                    variables=cat_features,
                    variables_type="categorical",
                    max_n_bins=10,
                )
            )
        if len(num_features) == 0 and len(cat_features) == 0:
            raise AssertionError("No numeric or categorical columns detected in X.")

        prebucketing_pipeline = to_skorecard_pipeline(make_pipeline(*prebucketing_pipe))
        bucketing_pipeline = to_skorecard_pipeline(make_pipeline(*bucketing_pipe))

        return BucketingProcess(
            specials=self.specials,
            prebucketing_pipeline=prebucketing_pipeline,
            bucketing_pipeline=bucketing_pipeline,
            random_state=self.random_state,
        )

    def _build_pipeline(self, X):
        """Build the default pipeline."""
        if self.encoder == "woe":
            # encoder = WoeEncoder()
            encoder = WOEEncoder(cols=list(X.columns))
        else:
            raise NotImplementedError(f"Encoder {self.encoder} not supported. Please use 'woe'")

        # If a pipeline is specified, convert to SkorecardPipeline
        if self.bucketing is None:
            self.bucketing_ = self._build_default_bucketing_process(X)
        elif isinstance(self.bucketing, Pipeline):
            for step in _get_all_steps(self.bucketing):
                if hasattr(step, "random_state") and self.random_state is not None:
                    if step.random_state is not None:
                        warnings.warn(f"Overwriting random_state of {step} with random_state of Skorecard", UserWarning)
                    step.random_state = self.random_state

            self.bucketing_ = to_skorecard_pipeline(self.bucketing)
        else:
            self.bucketing_ = self.bucketing
            if isinstance(self.bucketing_, BucketingProcess):
                self.bucketing_.random_state = self.random_state

        # Note ColumnSelector will not select any columns if passed an empty list.
        self.pipeline_ = Pipeline(
            [
                ("bucketer", self.bucketing_),
                ("encoder", encoder),
                ("column_selector", ColumnSelector(self.variables)),
                ("model", LogisticRegression(calculate_stats=self.calculate_stats, **self.lr_kwargs)),
            ]
        )

    def fit(self, X, y=None):
        """Fit the skorecard pipeline with X, y.

        Args:
            X (pd.DataFrame): features
            y (pd.Series, optional): target. Defaults to None.
        """
        # input validation
        assert isinstance(self.variables, list)
        # data validation
        X = ensure_dataframe(X)
        self.classes_, y = np.unique(y, return_inverse=True)

        self._build_pipeline(X)

        try:
            self.pipeline_.fit(X, y)
        except np.linalg.LinAlgError as e:
            # update the LinAlgError message with a more helpful message.
            error_msg = (
                "\nThe LinAlgError is very likely caused by multi-collinear variables."
                "Check for multi-collinearity by calling \n"
                "model.woe_transform(X).corr()"
                "\nThis is probably due to at least two features being bucketed to one unique bucket.\n"
                "Find those features by running\n"
                "model.bucket_transform(X_train).nunique()"
            )
            raise type(e)(f"{e.args[0]}\n{error_msg}")

        # Save some stuff for scikitlearn
        self.coef_ = self.pipeline_[-1].coef_
        self.n_features_in_ = len(X.columns)

        check_suppressor_effect(self.coef_[0], X.columns)

        return self

    def fit_interactive(self, X, y=None, mode="external", **server_kwargs):
        """
        Fit a bucketer and then interactive edit the fit using a dash app.

        Note we are using a [jupyterdash](https://medium.com/plotly/introducing-jupyterdash-811f1f57c02e) app,
        which supports 3 different modes:

        - 'external' (default): Start dash server and print URL
        - 'inline': Start dash app inside an Iframe in the jupyter notebook
        - 'jupyterlab': Start dash app as a new tab inside jupyterlab

        """
        # We need to make sure we only fit if not already fitted
        # This prevents a user losing manually defined boundaries
        # when re-running .fit_interactive()
        if not is_fitted(self):
            self.fit(X, y)
        self.bucketing_.fit_interactive(X=X, y=y, mode=mode, **server_kwargs)

    def bucket_transform(self, X):
        """
        Transform X through the bucketing pipelines.
        """
        check_is_fitted(self)
        return self.pipeline_[0].transform(X)

    def woe_transform(self, X):
        """Transform X through the bucketing + WoE pipelines."""
        check_is_fitted(self)
        return self.pipeline_[:2].transform(X)

    def predict_proba(self, X):
        """Predicted probabilities."""
        check_is_fitted(self)
        return self.pipeline_.predict_proba(X)

    def predict(self, X):
        """Predict the class."""
        check_is_fitted(self)
        return self.pipeline_.predict(X)

    def decision_function(self, *args, **kwargs):
        """Decision function."""
        check_is_fitted(self)
        return self.pipeline_.decision_function(*args, **kwargs)

    def get_stats(self):
        """Get the stats of the fitted model."""
        check_is_fitted(self)
        # Get the pipeline model
        model = self.pipeline_[-1]
        check_is_fitted(model)
        return model.get_stats()

    def summary(self):
        """Get the summary of all the bucketers."""
        raise NotImplementedError("Not implemented yet")
        check_is_fitted(self)
        return self.bucketing_.summary()

    def prebucket_table(self, column):
        """Get the prebucket_table.

        It's supported only if the bucketing object is BucketingProcess.

        Args:
            column: (str). column name

        Returns: pd.DataFrame, pre-bucketing summary for columns
        """
        check_is_fitted(self)
        if isinstance(self.bucketing_, BucketingProcess):
            return self.bucketing_.prebucket_table(column)
        else:
            error_msg = (
                f"prebucket_table is supported only if the attribute bucketing is of type BucketingProcess, "
                f"got {self.bucketing_.__class__}"
            )
            raise BucketerTypeError(error_msg)

    def bucket_table(self, column):
        """Get the bucket table.

        Args:
            column: (str). column name

        Returns: pd.DataFrame, -bucketing summary for columns
        """
        check_is_fitted(self)
        return self.bucketing_.bucket_table(column)

    def plot_prebucket(self, column, line="event_rate"):
        """Plot the prebucket_table. It's supported only if the bucketing object is BucketingProcess.

        Args:
            column: (str). column name
            line (str): The line to plot on the y-axis. Can be event rate or weight of evidence
        """
        if isinstance(self.bucketing_, BucketingProcess):
            return self.bucketing_.plot_prebucket(column, line=line)
        else:
            error_msg = (
                f"plot_prebucket  is supported only if the attribute bucketing is of type BucketingProcess, "
                f"got {type(self.bucketing_.__class__)}"
            )
            raise BucketerTypeError(error_msg)

    def plot_bucket(self, column, line="event_rate"):
        """Plot the buckets.

        Args:
            column (str): column name
            line (str): The line to plot on the y-axis. Can be event rate or weight of evidence
        """
        check_is_fitted(self)
        return self.bucketing_.plot_bucket(column, line=line)

    def _more_tags(self):
        """
        Estimator tags are annotations of estimators that allow programmatic inspection of their capabilities.

        See https://scikit-learn.org/stable/developers/develop.html#estimator-tags
        """  # noqa
        return {"binary_only": True, "allow_nan": True}



import pytest
from skorecard import datasets


def test_load_data_as_frame():
    """Test setting as_frame to True."""
    df = datasets.load_uci_credit_card(return_X_y=False, as_frame=True)
    assert df.shape == (6000, 5)


def test_load_data_as_frame_and_X_y():
    """Test setting both params to True gives an error."""
    with pytest.raises(ValueError):
        X, y = datasets.load_uci_credit_card(return_X_y=True, as_frame=True)


def test_load_data_X_y():
    """Test we get X and y numpy arrays."""
    X, y = datasets.load_uci_credit_card(return_X_y=True, as_frame=False)
    assert X.shape == (6000, 4)
    assert y.shape == (6000,)


def test_load_data_dict():
    """Test that setting both arguments to False returns a dict."""
    X_y_dict = datasets.load_uci_credit_card(return_X_y=False, as_frame=False)
    assert isinstance(X_y_dict, dict)
    assert X_y_dict["data"].shape == (6000, 4)
    assert X_y_dict["target"].shape == (6000,)



import warnings
from typing import Dict

import pandas as pd
from sklearn.utils import check_array


def is_fitted(estimator) -> bool:
    """
    Checks if an estimator is fitted.

    Loosely taken from
    https://github.com/scikit-learn/scikit-learn/blob/2beed5584/sklearn/utils/validation.py#L1034
    """  # noqa

    if not hasattr(estimator, "fit"):
        raise TypeError("%s is not an estimator instance." % (estimator))

    attrs = [v for v in vars(estimator) if v.endswith("_") and not v.startswith("__")]

    return len(attrs) > 0


def ensure_dataframe(X: pd.DataFrame) -> pd.DataFrame:
    """
    Make sure X is a pandas DataFrame.
    """
    # checks if the input is a dataframe.
    if not isinstance(X, pd.DataFrame):
        # Convert X to pd.DataFrame. Not recommended,
        # as you'll lose column name info.
        # but bucketer will still work on numpy matrix
        # also required for full scikitlearn compatibility
        X = X.copy()
        X = check_array(X, force_all_finite=False, accept_sparse=False, dtype=None)
        X = pd.DataFrame(X)
        X.columns = list(X.columns)  # sometimes columns can be a RangeIndex..
    else:
        # Create a copy
        # important not to transform the original dataset.
        X = X.copy()

    if X.shape[0] == 0:
        raise ValueError("Dataset has no rows!")
    if X.shape[1] == 0:
        raise ValueError(f"0 feature(s) (shape=({X.shape[0]}, 0)) while a minimum of 1 is required.")

    return X


def check_args(args: Dict, obj):
    """
    Checks if keys from args dictionary are valid args to an object.

    Note: this assumes 'obj' is scikit-learn compatible and thus has .get_params() implemented.
    """
    valid_args = obj().get_params()
    for arg in args.keys():
        if arg not in valid_args:
            msg = f"Argument '{arg}' is not a valid argument for object '{obj}'"
            warnings.warn(msg)


def check_suppressor_effect(coefs: list, feat_names: list):
    """Checks if the coefficients all have the expected sign."""
    suspect_feats = []
    for i, c in enumerate(coefs):
        if c < 0:
            suspect_feats.append(feat_names[i])
    if len(suspect_feats) != 0:
        msg = f"Features found with coefficient-sign that is contrary to what is expected based on weight-of-evidence. \
                This is likely caused by multi-collinearity. The features are: {suspect_feats}"
        warnings.warn(msg)



from typing import Dict, List, Union

import numpy as np
import pandas as pd

from skorecard.bucket_mapping import BucketMapping
from skorecard.reporting.plotting import get_bucket_color


def determine_boundaries(df: pd.DataFrame, bucket_mapping: BucketMapping) -> Union[List, Dict]:
    """
    Determine mapping boundaries.

    Given a dataframe with pre_bucket and bucket column, determine the boundaries
    that can be passed to the bucket_mapping.

    ```python
    import pandas as pd
    from skorecard.bucket_mapping import BucketMapping
    df = pd.DataFrame()
    df['pre_buckets'] = [0,1,2,3,4,5,6,7,8,9,10]
    df['buckets'] = [0,0,1,1,2,2,2,3,3,4,5]

    bucket_mapping = BucketMapping('feature1', 'numerical', map = [2,3,4,5])

    determine_boundaries(df, bucket_mapping)
    ```
    """
    assert "pre_buckets" in df.columns
    assert "buckets" in df.columns

    # filter out specials
    df = df[df["buckets"] >= 0]

    # Categoricals are an easy 1 on 1 relationship
    if bucket_mapping.type == "categorical":
        return dict(zip(df["pre_buckets"].values, df["buckets"].values))

    # For numerical, for have right = True
    # We can simply take the max pre-bucket for each bucket
    dfg = df.groupby(["buckets"]).agg(["max"])
    dfg.columns = dfg.columns.get_level_values(1)
    boundaries = dfg["max"]

    # For numerical with right = False
    # because the prebuckets are integers
    # we can safely add 1 to make sure the
    # map includes the right prebuckets
    if bucket_mapping.right is False:
        boundaries += 1

    # Drop the last value,
    # This makes sure outlier values are in the same bucket
    # instead of a new one
    boundaries = list(boundaries)[:-1]

    return boundaries


def perc_data_bars(column):
    """
    Display bar plots inside a dash DataTable cell.

    Assumes a value between 0 - 100.

    Adapted from: https://dash.plotly.com/datatable/conditional-formatting
    """
    n_bins = 100
    bounds = [i * (1.0 / n_bins) for i in range(n_bins + 1)]
    ranges = [float(x) for x in range(101)]
    styles = []
    for i in range(1, len(bounds)):
        min_bound = ranges[i - 1]
        max_bound = ranges[i]
        max_bound_percentage = bounds[i] * 100
        # For odd rows
        styles.append(
            {
                "if": {
                    "filter_query": (
                        "{{{column}}} >= {min_bound}"
                        + (" && {{{column}}} < {max_bound}" if (i < len(bounds) - 1) else "")
                    ).format(column=column, min_bound=min_bound, max_bound=max_bound),
                    "column_id": column,
                    "row_index": "odd",
                },
                "background": (
                    """
                    linear-gradient(90deg,
                    #0074D9 0%,
                    #0074D9 {max_bound_percentage}%,
                    rgb(248, 248, 248) {max_bound_percentage}%,
                    rgb(248, 248, 248) 100%)
                """.format(max_bound_percentage=max_bound_percentage)
                ),
                "paddingBottom": 2,
                "paddingTop": 2,
            }
        )
        # For even rows
        styles.append(
            {
                "if": {
                    "filter_query": (
                        "{{{column}}} >= {min_bound}"
                        + (" && {{{column}}} < {max_bound}" if (i < len(bounds) - 1) else "")
                    ).format(column=column, min_bound=min_bound, max_bound=max_bound),
                    "column_id": column,
                    "row_index": "even",
                },
                "background": (
                    """
                    linear-gradient(90deg,
                    #0074D9 0%,
                    #0074D9 {max_bound_percentage}%,
                    white {max_bound_percentage}%,
                    white 100%)
                """.format(max_bound_percentage=max_bound_percentage)
                ),
                "paddingBottom": 2,
                "paddingTop": 2,
            }
        )

    return styles


def colorize_cell(column):
    """Colourize the integer bucket number.

    We can safely assume max 20 buckets, as features are often binned to 3-7 buckets.
    We will cycle through them.
    """
    styles = []
    for i in range(-10, 21):
        styles.append(
            {
                "if": {
                    # 'row_index': i,  # number | 'odd' | 'even'
                    "filter_query": f"{{{column}}} = '{i}'",
                    "column_id": column,
                },
                "backgroundColor": get_bucket_color(i),
                "color": "white",
            }
        )
    return styles


def is_monotonic_increasing(x):
    """
    Helper function to determine if a list is monotonically increasing.
    """
    dx = np.diff(x)
    return np.all(dx >= 0)


def is_increasing(x):
    """
    Helper function to determine if a list is increasing.
    """
    dx = np.diff(x)
    return np.all(dx > 0)


def is_sequential(x):
    """
    Helper function to determine if a list is monotonically increasing with step size 1.
    """
    dx = np.diff(x)
    return np.all(np.isin(dx, [0, 1]))



"""Import required transformers."""

from .bucketers import (
    OptimalBucketer,
    EqualWidthBucketer,
    EqualFrequencyBucketer,
    AgglomerativeClusteringBucketer,
    DecisionTreeBucketer,
    OrdinalCategoricalBucketer,
    AsIsNumericalBucketer,
    AsIsCategoricalBucketer,
    UserInputBucketer,
)

__all__ = [
    "OptimalBucketer",
    "EqualWidthBucketer",
    "AgglomerativeClusteringBucketer",
    "EqualFrequencyBucketer",
    "DecisionTreeBucketer",
    "OrdinalCategoricalBucketer",
    "AsIsNumericalBucketer",
    "AsIsCategoricalBucketer",
    "UserInputBucketer",
]



from typing import List, Tuple

import pandas as pd

from skorecard.utils import NotInstalledError

try:
    import dabl as db
except ModuleNotFoundError:
    db = NotInstalledError("dabl")


def detect_types(X: pd.DataFrame) -> Tuple[List, List]:
    """Detects numerical and categorical columns.

    Wrapper around `dabl.detect_types`.

    ```python
    from skorecard import datasets
    from skorecard.utils import detect_types

    X, y = datasets.load_uci_credit_card(return_X_y=True)

    num_cols, cat_cols = detect_types(X)
    print(f"Categorical column(s) = {cat_cols}")
    print(f"Numerical column(s) = {num_cols}")
    ```
    Args:
        X (pd.DataFrame): Input dataframe
    """
    assert isinstance(X, pd.DataFrame)
    detected_types = db.detect_types(X)

    cat_columns = X.columns[(detected_types["categorical"] is True) | (detected_types["low_card_int"] is True)]
    num_columns = X.columns[(detected_types["continuous"] is True) | (detected_types["dirty_float"] is True)]

    return list(num_columns), list(cat_columns)



import numpy as np
import pandas as pd
import scipy
from sklearn import linear_model as lm
from sklearn.utils.validation import check_is_fitted

from skorecard.reporting import weight_plot
from skorecard.utils import convert_sparse_matrix


class LogisticRegression(lm.LogisticRegression):
    """Extended Logistic Regression.

    Extends [sklearn.linear_model.LogisticRegression](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html).

    This class provides the following extra statistics, calculated on `.fit()` and accessible via `.get_stats()`:

    - `cov_matrix_`: covariance matrix for the estimated parameters.
    - `std_err_intercept_`: estimated uncertainty for the intercept
    - `std_err_coef_`: estimated uncertainty for the coefficients
    - `z_intercept_`: estimated z-statistic for the intercept
    - `z_coef_`: estimated z-statistic for the coefficients
    - `p_value_intercept_`: estimated p-value for the intercept
    - `p_value_coef_`: estimated p-value for the coefficients

    Example:

    ```python
    from skorecard.datasets import load_uci_credit_card
    from skorecard.bucketers import EqualFrequencyBucketer
    from skorecard.linear_model import LogisticRegression
    from sklearn.pipeline import Pipeline
    from sklearn.preprocessing import OneHotEncoder

    X, y = load_uci_credit_card(return_X_y=True)

    pipeline = Pipeline([
        ('bucketer', EqualFrequencyBucketer(n_bins=10)),
        ('clf', LogisticRegression(calculate_stats=True))
    ])
    pipeline.fit(X, y)
    assert pipeline.named_steps['clf'].p_val_coef_[0][0] > 0

    pipeline.named_steps['clf'].get_stats()
    ```

    An example output of `.get_stats()`:

    Index     | Coef.     | Std.Err  |   z       | Pz
    --------- | ----------| ---------| ----------| ------------
    const     | -0.537571 | 0.096108 | -5.593394 | 2.226735e-08
    EDUCATION | 0.010091  | 0.044874 | 0.224876  | 8.220757e-01

    """  # noqa

    def __init__(
        self,
        penalty="l2",
        calculate_stats=False,
        dual=False,
        tol=0.0001,
        C=1.0,
        fit_intercept=True,
        intercept_scaling=1,
        class_weight=None,
        random_state=None,
        solver="lbfgs",
        max_iter=100,
        multi_class="auto",
        verbose=0,
        warm_start=False,
        n_jobs=None,
        l1_ratio=None,
    ):
        """
        Extends [sklearn.linear_model.LogisticRegression.fit()](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html).

        Args:
            calculate_stats (bool): If true, calculate statistics like standard error during fit, accessible with .get_stats()
        """  # noqa
        super().__init__(
            penalty=penalty,
            dual=dual,
            tol=tol,
            C=C,
            fit_intercept=fit_intercept,
            intercept_scaling=intercept_scaling,
            class_weight=class_weight,
            random_state=random_state,
            solver=solver,
            max_iter=max_iter,
            multi_class=multi_class,
            verbose=verbose,
            warm_start=warm_start,
            n_jobs=n_jobs,
            l1_ratio=l1_ratio,
        )
        self.calculate_stats = calculate_stats

    def fit(self, X, y, sample_weight=None, calculate_stats=False, **kwargs):
        """
        Fit the model.

        Overwrites [sklearn.linear_model.LogisticRegression.fit()](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html).

        In addition to the standard fit by sklearn, this function will compute the covariance of the coefficients.

        Args:
            X (array-like, sparse matrix): Matrix of shape (n_samples, n_features)
                Training vector, where n_samples is the number of samples and
                n_features is the number of features.
            y (array-like): of shape (n_samples,)
                Target vector relative to X.
            sample_weight (array-like): of shape (n_samples,) default=None
                Array of weights that are assigned to individual samples.
                If not provided, then each sample is given unit weight.
            calculate_stats (bool): If true, calculate statistics like standard error during fit, accessible with .get_stats()

        Returns:
            self (LogisticRegression): Fitted estimator.
        """  # noqa

        if not self.calculate_stats and not calculate_stats:
            return super().fit(X, y, sample_weight=sample_weight, **kwargs)

        X = convert_sparse_matrix(X)
        if isinstance(X, pd.DataFrame):
            self.names_ = ["const"] + [f for f in X.columns]
        else:
            self.names_ = ["const"] + [f"x{i}" for i in range(X.shape[1])]

        lr = super().fit(X, y, sample_weight=sample_weight, **kwargs)

        predProbs = self.predict_proba(X)

        # Design matrix -- add column of 1's at the beginning of your X matrix
        if lr.fit_intercept:
            X_design = np.hstack([np.ones((X.shape[0], 1)), X])
        else:
            X_design = X

        p = np.prod(predProbs, axis=1)
        self.cov_matrix_ = np.linalg.inv((X_design * p[..., np.newaxis]).T @ X_design)
        std_err = np.sqrt(np.diag(self.cov_matrix_)).reshape(1, -1)

        # In case fit_intercept is set to True, then in the std_error array
        # Index 0 corresponds to the intercept, from index 1 onwards it relates to the coefficients
        # If fit intercept is False, then all the values are related to the coefficients
        if lr.fit_intercept:
            self.std_err_intercept_ = std_err[:, 0]
            self.std_err_coef_ = std_err[:, 1:][0]

            self.z_intercept_ = self.intercept_ / self.std_err_intercept_

            # Get p-values under the gaussian assumption
            self.p_val_intercept_ = scipy.stats.norm.sf(abs(self.z_intercept_)) * 2

        else:
            self.std_err_intercept_ = np.array([np.nan])
            self.std_err_coef_ = std_err[0]

            self.z_intercept_ = np.array([np.nan])

            # Get p-values under the gaussian assumption
            self.p_val_intercept_ = np.array([np.nan])

        self.z_coef_ = self.coef_ / self.std_err_coef_
        self.p_val_coef_ = scipy.stats.norm.sf(abs(self.z_coef_)) * 2

        return self

    def get_stats(self) -> pd.DataFrame:
        """
        Puts the summary statistics of the fit() function into a pandas DataFrame.

        Returns:
            data (pandas DataFrame): The statistics dataframe, indexed by
                the column name
        """
        check_is_fitted(self)

        if not hasattr(self, "std_err_coef_"):
            msg = "Summary statistics were not calculated on .fit(). Options to fix:\n"
            msg += "\t- Re-fit using .fit(X, y, calculate_stats=True)\n"
            msg += "\t- Re-inititialize using LogisticRegression(calculate_stats=True)"
            raise AssertionError(msg)

        data = {
            "Coef.": (self.intercept_.tolist() + self.coef_.tolist()[0]),
            "Std.Err": (self.std_err_intercept_.tolist() + self.std_err_coef_.tolist()),
            "z": (self.z_intercept_.tolist() + self.z_coef_.tolist()[0]),
            "P>|z|": (self.p_val_intercept_.tolist() + self.p_val_coef_.tolist()[0]),
        }

        return pd.DataFrame(data, index=self.names_)

    def plot_weights(self):
        """
        Plots the relative importance of coefficients of the model.

        Example:

        ```from skorecard.datasets import load_uci_credit_card
        from skorecard.bucketers import EqualFrequencyBucketer
        from skorecard.linear_model import LogisticRegression
        from skorecard.reporting.plotting import weight_plot
        from sklearn.pipeline import Pipeline
        from sklearn.preprocessing import OneHotEncoder
        X, y = load_uci_credit_card(return_X_y=True)
        pipeline = Pipeline([
            ('bucketer', EqualFrequencyBucketer(n_bins=10)),
            ('clf', LogisticRegression(calculate_stats=True))
        ])
        pipeline.fit(X, y)
        assert pipeline.named_steps['clf'].p_val_coef_[0][0] > 0
        stats = pipeline.named_steps['clf'].get_stats()
        pipeline.named_steps['clf'].plot_weights()```
        """
        stats = self.get_stats()
        return weight_plot(stats)



from .arrayfuncs import (
    reshape_1d_to_2d,
    convert_sparse_matrix,
)
from .exceptions import (
    DimensionalityError,
    UnknownCategoryError,
    NotInstalledError,
    NotBucketedError,
    NotPreBucketedError,
    NotBucketObjectError,
    BucketingPipelineError,
    BucketerTypeError,
)
from .dataframe import detect_types


__all__ = [
    "reshape_1d_to_2d",
    "convert_sparse_matrix",
    "DimensionalityError",
    "UnknownCategoryError",
    "NotInstalledError",
    "NotBucketObjectError",
    "detect_types",
    "NotBucketedError",
    "NotPreBucketedError",
    "BucketingPipelineError",
    "BucketerTypeError",
]



class NotPreBucketedError(Exception):
    """Raise this when a feature is not prebucketed (has >100 unique values)."""

    def __init__(self, message):
        """Raise this when this when a feature is not prebucketed (has >100 unique values).

        Args:
            message: (str) message when exception is raised
        """
        self.message = message


class NotBucketedError(Exception):
    """Raise this when a feature is not Bucketed but should have been."""

    def __init__(self, message):
        """Raise this when a feature is not Bucketed but should have been.

        Args:
            message: (str) message when exception is raised
        """
        self.message = message


class NotBucketObjectError(Exception):
    """Raise this when a non-bucket object is found in the bucketing_process."""

    def __init__(self, message):
        """Raise this when a non-bucket object is found in the bucketing_process.

        Args:
            message: (str) message when exception is raised
        """
        self.message = message


class BucketingPipelineError(Exception):
    """Raise this when there is something wrong with sklearn pipeline with bucketing inside."""

    def __init__(self, message):
        """Raise this when there is something wrong with sklearn pipeline with bucketing inside.

        Args:
            message: (str) message when exception is raised
        """
        self.message = message


class DimensionalityError(Exception):
    """Raise this when the Dimension of the numpy array or pandas DataFrame is wrong."""

    def __init__(self, message):
        """Raise this when the Dimension of the numpy array or pandas DataFrame is wrong.

        Args:
            message: (str) message when exception is raised
        """
        self.message = message


class UnknownCategoryError(Exception):
    """Raise this when an array contains a new unseen category."""

    def __init__(self, message):
        """Raise this when array contains a new unseen category.

        Args:
            message: (str) message when exception is raised
        """
        self.message = message


class NotInstalledError:
    """
    This object is used for optional dependencies.

    This allows us to give a friendly message to the user that they need to install extra dependencies as well as a link
    to our documentation page.

    Adapted from: https://github.com/RasaHQ/whatlies/blob/master/whatlies/error.py

    Example usage:

    ```python
    from skorecard.utils import NotInstalledError

    try:
        import dash_core_components as dcc
    except ModuleNotFoundError as e:
        dcc = NotInstalled("dash_core_components", "dashboard")

    dcc.Markdown() # Will raise friendly error with instructions how to solve
    ```

    Note that installing optional dependencies in a package are defined in setup.py.
    """

    def __init__(self, tool, dep=None):
        """Initialize error with missing package and reference to conditional install package.

        Args:
            tool (str): The name of the pypi package that is missing
            dep (str): The name of the extra_imports set (defined in setup.py) where the package is present. (optional)
        """
        self.tool = tool
        self.dep = dep

        msg = f"In order to use {self.tool} you'll need to install via;\n\n"
        if self.dep is None:
            msg += f"pip install {self.tool}\n\n"
        else:
            msg += f"pip install skorecard[{self.dep}]\n\n"

        self.msg = msg

    def __getattr__(self, *args, **kwargs):
        """Raise when accessing an attribute."""
        raise ModuleNotFoundError(self.msg)

    def __call__(self, *args, **kwargs):
        """Raise when accessing a method."""
        raise ModuleNotFoundError(self.msg)


class ApproximationWarning(Warning):
    """
    Warning for approx.
    """

    def __init__(self, message):
        """
        Init.
        """
        self.message = message


class BucketerTypeError(Exception):
    """Raise this when there is something wrong with Bucketer."""

    def __init__(self, message):
        """Raise this when there is something wrong with sklearn pipeline with bucketing inside.

        Args:
            message: (str) message when exception is raised
        """
        self.message = message



import pandas as pd
import pytest
from sklearn.pipeline import make_pipeline

from skorecard.bucketers import DecisionTreeBucketer, OrdinalCategoricalBucketer
from skorecard.linear_model import LogisticRegression
from skorecard.preprocessing import ColumnSelector, WoeEncoder


def test_column_selector(df):
    """Test that the column selector works."""
    features = ["LIMIT_BAL", "BILL_AMT1"]
    X = df
    # y = df["default"]

    cs = ColumnSelector(variables=features)
    X_sel = cs.fit_transform(X)

    assert isinstance(X_sel, pd.DataFrame)
    assert X_sel.shape[1] == len(features)

    # test that it raises an exception if X is not a pd DataFrame
    npX = X.values
    cs = ColumnSelector(variables=features)
    with pytest.raises(KeyError):
        cs.fit_transform(npX)


def test_column_selector_in_pipeline(df):
    """Test that the column selector works well in a pipeline."""
    features = ["LIMIT_BAL", "BILL_AMT1"]
    X = df.drop("default", axis=1)
    y = df["default"]

    model_pipe = make_pipeline(
        DecisionTreeBucketer(variables=["LIMIT_BAL"], max_n_bins=5),
        OrdinalCategoricalBucketer(variables=["MARRIAGE"], tol=0.05),
        WoeEncoder(),
        ColumnSelector(variables=features),
        LogisticRegression(calculate_stats=True),
    ).fit(X, y)

    # test that the model pipe has as many coefficients as there are features.
    assert len(model_pipe[-1].coef_[0]) == len(features)

    # test that the stats of the model pipe represent the correct features with a selector.
    model_stats = model_pipe[-1].get_stats().index.tolist()
    model_stats.remove("const")
    assert model_stats == features
